from itertools import chain, product

from lalamo.model_import.decoder_configs import HFLFM2Config
from lalamo.models.language_model import GenerationConfig
from lalamo.quantization import QuantizationMode

from .common import ConfigMap, FileSpec, ModelSpec

__all__ = ["LFM2_MODELS"]


def _lfm_repo(family: str, size: str, variant: str | None, quantization: QuantizationMode | None) -> tuple[str, str]:
    return (
        "LiquidAI" if quantization is None else "mlx-community",
        f"{family}-{size}"
        f"{f'-{variant}' if variant is not None else ''}"
        f"{f'-{quantization.bits}bit' if quantization is not None else ''}",
    )


_LFM20_MODELS = [
    ModelSpec(
        vendor="LiquidAI",
        family="LFM2",
        name=_lfm_repo("LFM2", size, variant, quantization)[1],
        size=size,
        repo="/".join(_lfm_repo("LFM2", size, variant, quantization)),
        config_type=HFLFM2Config,
        quantization=quantization,
        configs=ConfigMap(
            generation_config=GenerationConfig(temperature=0.3, min_p=0.15),  # , repetition_penalty=1.05
            chat_template=FileSpec("chat_template.jinja"),
        ),
        use_cases=tuple(),
    )
    for size, variant, quantization in chain(
        product(["350M", "700M", "1.2B"], [None], [None, QuantizationMode.UINT4, QuantizationMode.UINT8]),
        product(["2.6B"], [None, "Exp"], [None]),
        product(["2.6B"], ["Exp"], [QuantizationMode.UINT4, QuantizationMode.UINT8]),
    )
]

_LFM25_MODELS = [
    ModelSpec(
        vendor="LiquidAI",
        family="LFM2.5",
        name=_lfm_repo("LFM2.5", size, variant, quantization)[1],
        size=size,
        repo="/".join(_lfm_repo("LFM2.5", size, variant, quantization)),
        config_type=HFLFM2Config,
        quantization=quantization,
        configs=ConfigMap(
            generation_config=GenerationConfig(temperature=0.1, top_k=50, top_p=0.1),  # , repetition_penalty=1.05
            chat_template=FileSpec("chat_template.jinja"),
        ),
        use_cases=tuple(),
    )
    for size, variant, quantization in chain(
        product(["1.2B"], ["Instruct"], [None]),
    )
]

LFM2_MODELS = _LFM20_MODELS + _LFM25_MODELS



from collections.abc import Mapping
from dataclasses import dataclass, replace
from enum import Enum
from typing import Self

import equinox as eqx
import jax
from jax import numpy as jnp
from jaxtyping import Array, DTypeLike, Float

from lalamo.common import ParameterTree, dummy_array

from .common import LalamoModule

__all__ = [
    "Normalization",
    "NormalizationConfig",
    "UpcastMode",
]


class UpcastMode(Enum):
    ONLY_NORMALIZATION = "only_normalization"
    FULL_LAYER = "full_layer"


@dataclass(frozen=True)
class NormalizationConfig:
    scale_precision: DTypeLike
    accumulation_precision: DTypeLike
    epsilon: float
    scale_offset: float | None
    upcast_mode: UpcastMode
    subtract_mean: bool

    def init(self, input_dim: int) -> "Normalization":
        scales = jnp.ones(input_dim, dtype=self.scale_precision)
        return Normalization(self, scales=scales)

    def empty(self, input_dim: int) -> "Normalization":
        return Normalization(
            config=self,
            scales=dummy_array(input_dim, dtype=self.scale_precision),
        )


class Normalization(LalamoModule[NormalizationConfig]):
    scales: Float[Array, " channels"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.scale_precision

    @property
    def input_dim(self) -> int:
        (result,) = self.scales.shape
        return result

    def __post_init__(self) -> None:
        if self.config.scale_precision != self.scales.dtype:
            raise ValueError(
                f"Scales precision {self.scales.dtype} does not match the"
                f" specified precision {self.config.scale_precision}",
            )

    @eqx.filter_jit
    def __call__(self, inputs: Float[Array, " channels"]) -> Float[Array, " channels"]:
        upcasted_inputs = inputs.astype(self.config.accumulation_precision)

        if self.config.subtract_mean:
            mean = jnp.mean(upcasted_inputs)
            upcasted_inputs = upcasted_inputs - mean

        adjusted_variance = jnp.mean(jnp.square(upcasted_inputs)) + self.config.epsilon
        normalized_x = upcasted_inputs * jax.lax.rsqrt(adjusted_variance)

        if self.config.upcast_mode == UpcastMode.ONLY_NORMALIZATION:
            normalized_x = normalized_x.astype(inputs.dtype)

        if self.config.upcast_mode == UpcastMode.FULL_LAYER:
            adjusted_scales = self.scales.astype(self.config.accumulation_precision)
        else:
            adjusted_scales = self.scales

        if self.config.scale_offset is not None:
            adjusted_scales = adjusted_scales + self.config.scale_offset

        result = normalized_x * adjusted_scales
        return result.astype(inputs.dtype)

    def export_weights(self) -> ParameterTree:
        return {"scales": self.scales}

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        return replace(self, scales=weights["scales"])



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Literal

from jaxtyping import DTypeLike

from lalamo.modules import (
    AttentionConfig,
    DecoderConfig,
    DenseMLPConfig,
    FullPrecisionLinearConfig,
    MixtureOfExpertsConfig,
    NormalizationConfig,
    SoftmaxRouting,
    TiedEmbeddingConfig,
    TransformerConfig,
    TransformerLayerConfig,
    UntiedEmbeddingConfig,
    UpcastMode,
    YARNRoPEConfig,
)
from lalamo.modules.activations import SiLU

from .common import HuggingFaceLMConfig

__all__ = ["HFGPTOssConfig"]


@dataclass(frozen=True)
class YarnRopeScalingConfig:
    factor: float
    beta_fast: float
    beta_slow: float
    original_max_position_embeddings: int
    rope_type: Literal["yarn"]
    truncate: bool


@dataclass(frozen=True)
class HFGPTOssConfig(HuggingFaceLMConfig):
    # Core HF fields
    architectures: list[Literal["GptOssForCausalLM"]]
    attention_bias: bool
    attention_dropout: float
    eos_token_id: int | list[int]
    hidden_act: Literal["silu"]
    hidden_size: int
    initializer_range: float
    intermediate_size: int
    max_position_embeddings: int
    model_type: Literal["gpt_oss"]
    num_attention_heads: int
    num_hidden_layers: int
    num_key_value_heads: int
    pad_token_id: int
    rms_norm_eps: float
    rope_theta: float
    tie_word_embeddings: bool
    transformers_version: str
    use_cache: bool
    vocab_size: int

    # GPT-OSS specifics
    layer_types: list[Literal["sliding_attention", "full_attention"]] | None
    sliding_window: int | None
    swiglu_limit: float
    head_dim: int | None
    num_local_experts: int
    num_experts_per_tok: int | None = None
    experts_per_token: int | None = None  # some configs may use this alias
    rope_scaling: YarnRopeScalingConfig | None = None
    output_router_logits: bool | None = None
    router_aux_loss_coef: float | None = None

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        metadata_dict: Mapping[str, str],  # noqa: ARG002
    ) -> DecoderConfig:
        # Embedding
        if self.tie_word_embeddings:
            embedding_config = TiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                precision=activation_precision,
            )
        else:
            embedding_config = UntiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                precision=activation_precision,
            )

        if self.rope_scaling is not None and self.rope_scaling.rope_type == "yarn":
            rope_config = YARNRoPEConfig(
                precision=activation_precision,
                base=self.rope_theta,
                max_sequence_length=context_length or self.max_position_embeddings,
                scaling_factor=self.rope_scaling.factor,
                original_context_length=self.rope_scaling.original_max_position_embeddings,
                beta_fast=self.rope_scaling.beta_fast,
                beta_slow=self.rope_scaling.beta_slow,
                truncate=self.rope_scaling.truncate,
            )
        else:
            rope_config = YARNRoPEConfig(
                precision=activation_precision,
                base=self.rope_theta,
                max_sequence_length=context_length or self.max_position_embeddings,
                scaling_factor=1.0,
                original_context_length=self.max_position_embeddings,
                beta_fast=32.0,
                beta_slow=1.0,
                truncate=True,
            )

        rmsnorm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.rms_norm_eps,
            scale_offset=None,
            upcast_mode=UpcastMode.FULL_LAYER,
            subtract_mean=False,
        )

        # Linear layers
        linear_config = FullPrecisionLinearConfig(precision=activation_precision)

        # Experts (MoE) scaffold
        # Router: linear with bias; Experts: DenseMLP with SiLU(alpha=1.702) and value/gate clipping
        experts_activation = SiLU(alpha=1.702)
        experts_config = DenseMLPConfig(
            linear_config=linear_config,
            activation=experts_activation,
            has_up_biases=True,
            has_down_biases=True,
            up_clipping=(-self.swiglu_limit + 1.0, self.swiglu_limit + 1.0),
            gate_clipping=(None, self.swiglu_limit),
        )
        moe_config = MixtureOfExpertsConfig(
            mixture_size=self.num_local_experts,
            num_experts_per_token=(self.num_experts_per_tok or self.experts_per_token or 1),
            routing_function=SoftmaxRouting(),
            router_config=linear_config,
            router_has_biases=True,
            expert_config=experts_config,
        )

        # Per-layer sliding-window
        if self.layer_types is not None and len(self.layer_types) == self.num_hidden_layers:
            sliding_window_sizes = [
                self.sliding_window if layer_type == "sliding_attention" else None for layer_type in self.layer_types
            ]
        else:
            # Fallback: apply the same sliding window to all layers if provided
            sliding_window_sizes = (
                [self.sliding_window] * self.num_hidden_layers
                if self.sliding_window is not None
                else [None] * self.num_hidden_layers
            )

        head_dim = self.head_dim if self.head_dim is not None else self.hidden_size // self.num_attention_heads

        layer_configs = []
        for sliding_window_size in sliding_window_sizes:
            attention_config = AttentionConfig(
                qkv_projection_config=linear_config,
                out_projection_config=linear_config,
                query_norm_config=None,
                key_norm_config=None,
                logit_soft_cap=None,
                has_sinks=True,
                has_qkv_biases=self.attention_bias,
                has_out_biases=self.attention_bias,
                num_heads=self.num_attention_heads,
                num_groups=self.num_key_value_heads,
                head_dim=head_dim,
                is_causal=True,
                scale=None,
                sliding_window_size=sliding_window_size,
            )
            transformer_layer_config = TransformerLayerConfig(
                pre_mixer_norm_config=rmsnorm_config,
                mixer_config=attention_config,
                post_mixer_norm_config=None,
                pre_mlp_norm_config=rmsnorm_config,
                mlp_config=moe_config,
                post_mlp_norm_config=None,
            )
            layer_configs.append(transformer_layer_config)

        transformer_config = TransformerConfig(
            global_rope_config=rope_config,
            local_rope_config=None,
            layer_configs=tuple(layer_configs),
            output_norm_config=rmsnorm_config,
            model_dim=self.hidden_size,
            hidden_dim=self.intermediate_size,
            context_length=context_length or self.max_position_embeddings,
        )

        return DecoderConfig(
            embedding_config=embedding_config,
            transformer_config=transformer_config,
            vocab_size=self.vocab_size,
        )



# Based on https://github.com/huggingface/transformers/blob/main/src/transformers/modeling_rope_utils.py
# Original PyTorch code copyright notice:
#
# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from collections.abc import Mapping
from dataclasses import dataclass, replace

import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, DTypeLike, Float, Int

from lalamo.common import ParameterTree

from .common import LalamoModule, register_config_union

__all__ = [
    "LinearScalingRoPEConfig",
    "LlamaRoPEConfig",
    "PositionalEmbeddings",
    "RoPE",
    "RoPEConfigBase",
    "UnscaledRoPEConfig",
    "YARNRoPEConfig",
]


class PositionalEmbeddings(eqx.Module):
    cosines: Float[Array, "*batch tokens head_channels"]
    sines: Float[Array, "*batch tokens head_channels"]

    @property
    def head_dim(self) -> int:
        return self.cosines.shape[-1]

    def rotate_half(
        self,
        heads: Float[Array, "*batch tokens head_channels"],
    ) -> Float[Array, "*batch tokens head_channels"]:
        x1 = heads[..., : self.head_dim // 2]
        x2 = heads[..., self.head_dim // 2 :]
        return jnp.concatenate((-x2, x1), axis=-1)

    def apply(self, heads: Float[Array, "*batch tokens head_channels"]) -> Float[Array, "*batch tokens head_channels"]:
        return heads * self.cosines + self.rotate_half(heads) * self.sines

    def export(self) -> ParameterTree:
        return dict(
            cosines=self.cosines,
            sines=self.sines,
        )


@dataclass(frozen=True)
class RoPEConfigBase:
    precision: DTypeLike
    base: float
    max_sequence_length: int

    @property
    def _attention_scaling_factor(self) -> float:
        return 1.0

    def _scale_inverse_frequencies(
        self,
        inverse_frequencies: Float[Array, " tokens"],
        head_dim: int,  # noqa: ARG002
        max_sequence_length: int,  # noqa: ARG002
    ) -> Float[Array, " tokens"]:
        return inverse_frequencies

    def init(
        self,
        head_dim: int,
        num_timesteps: int,
    ) -> "RoPE":
        timesteps = jnp.arange(num_timesteps, dtype=jnp.float32)
        channel_indices = jnp.arange(0, head_dim, 2, dtype=jnp.int32)
        inverse_frequencies = 1.0 / (self.base ** (channel_indices.astype(jnp.float32) / head_dim))
        inverse_frequencies = self._scale_inverse_frequencies(inverse_frequencies, head_dim, self.max_sequence_length)
        outer_inverse_frequencies = jnp.outer(timesteps, inverse_frequencies)
        embeddings = jnp.concatenate((outer_inverse_frequencies, outer_inverse_frequencies), axis=-1)
        cosines = (jnp.cos(embeddings) * self._attention_scaling_factor).astype(self.precision)
        sines = (jnp.sin(embeddings) * self._attention_scaling_factor).astype(self.precision)
        return RoPE(config=self, cosines=cosines, sines=sines)


class RoPE(LalamoModule[RoPEConfigBase]):
    sines: Float[Array, "tokens head_channels"]
    cosines: Float[Array, "tokens head_channels"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    def __post_init__(self) -> None:
        num_tokens, _ = self.sines.shape
        if num_tokens > self.config.max_sequence_length:
            raise ValueError(
                f"{num_tokens} exceeds the specified max sequence length {self.config.max_sequence_length}",
            )
        if self.cosines.dtype != self.config.precision:
            raise ValueError(
                f"Cosines dtype {self.cosines.dtype} does not match the specified precision {self.config.precision}",
            )
        if self.sines.dtype != self.config.precision:
            raise ValueError(
                f"Sines dtype {self.sines.dtype} does not match the specified precision {self.config.precision}",
            )
        if self.cosines.shape != self.sines.shape:
            raise ValueError(
                f"Cosines and sines shape mismatch: cosines have shape {self.cosines.shape},"
                f" while sines have shape {self.sines.shape}",
            )

    @property
    def head_dim(self) -> int:
        _, result = self.sines.shape
        return result

    @property
    def max_sequence_length(self) -> int:
        result, _ = self.sines.shape
        return result

    @eqx.filter_jit
    def __call__(self, timesteps: Int[Array, " tokens"]) -> PositionalEmbeddings:
        return PositionalEmbeddings(
            cosines=self.cosines[timesteps],
            sines=self.sines[timesteps],
        )

    def export_weights(self) -> ParameterTree[Array]:
        return {
            "cosines": self.cosines,
            "sines": self.sines,
        }

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> "RoPE":
        assert isinstance(weights, Mapping)
        return replace(self, cosines=weights["cosines"], sines=weights["sines"])


class UnscaledRoPEConfig(RoPEConfigBase):
    pass


@dataclass(frozen=True)
class LlamaRoPEConfig(RoPEConfigBase):
    scaling_factor: float
    original_context_length: int
    low_frequency_factor: float
    high_frequency_factor: float

    def _scale_inverse_frequencies(
        self,
        inverse_frequencies: Float[Array, " tokens"],
        head_dim: int,  # noqa: ARG002
        max_sequence_length: int,  # noqa: ARG002
    ) -> Float[Array, " tokens"]:
        low_frequency_wavelength = self.original_context_length / self.low_frequency_factor
        high_frequency_wavelength = self.original_context_length / self.high_frequency_factor

        wavelengths = 2 * math.pi / inverse_frequencies

        high_frequency_mask = wavelengths < high_frequency_wavelength
        low_frequency_mask = wavelengths > low_frequency_wavelength
        mid_frequency_mask = (~high_frequency_mask) & (~low_frequency_mask)

        smoothing_factors = self.original_context_length / wavelengths - self.low_frequency_factor
        smoothing_factors = smoothing_factors / (self.high_frequency_factor - self.low_frequency_factor)

        scaled_frequencies = inverse_frequencies / self.scaling_factor
        smoothly_scaled_frequencies = (
            smoothing_factors * inverse_frequencies + (1 - smoothing_factors) * scaled_frequencies
        )

        result = inverse_frequencies * high_frequency_mask.astype(jnp.float32)
        result = result + smoothly_scaled_frequencies * mid_frequency_mask.astype(jnp.float32)
        result = result + scaled_frequencies * low_frequency_mask.astype(jnp.float32)

        return result


@dataclass(frozen=True)
class YARNRoPEConfig(RoPEConfigBase):
    scaling_factor: float
    original_context_length: int
    beta_fast: float
    beta_slow: float
    truncate: bool

    @classmethod
    def _find_correction_dim(cls, num_rotations: float, dim: int, base: float, original_context_length: int) -> float:
        """Inverse dimension formula to find the dimension based on the number of rotations"""
        return (dim * math.log(original_context_length / (num_rotations * 2 * math.pi))) / (2 * math.log(base))

    @classmethod
    def _find_correction_range(
        cls,
        low_rot: float,
        high_rot: float,
        dim: int,
        base: float,
        original_context_length: int,
        truncate: bool,
    ) -> tuple[float, float]:
        """Find dimension range bounds based on rotations"""
        low = cls._find_correction_dim(low_rot, dim, base, original_context_length)
        high = cls._find_correction_dim(high_rot, dim, base, original_context_length)
        if truncate:
            low = math.floor(low)
            high = math.ceil(high)
        return max(low, 0.0), min(high, float(dim - 1))

    @classmethod
    def _linear_ramp_factor(cls, min_value: float, max_value: float, dim: int) -> Float[Array, " head_dim"]:
        if min_value == max_value:
            max_value += 0.001  # Prevent singularity

        min_v = jnp.float32(min_value)
        max_v = jnp.float32(max_value)
        linear_func = (jnp.arange(dim, dtype=jnp.float32) - min_v) / (max_v - min_v)
        ramp_func = jnp.clip(linear_func, 0, 1)
        return ramp_func

    def _scale_inverse_frequencies(
        self,
        inverse_frequencies: Float[Array, " tokens"],
        head_dim: int,
        max_sequence_length: int,  # noqa: ARG002
    ) -> Float[Array, " tokens"]:
        scaled_frequencies = inverse_frequencies / self.scaling_factor

        low, high = self._find_correction_range(
            self.beta_fast,
            self.beta_slow,
            head_dim,
            self.base,
            self.original_context_length,
            self.truncate,
        )

        # Get n-dimensional rotational scaling corrected for extrapolation
        smoothing_factor = 1 - self._linear_ramp_factor(low, high, head_dim // 2)
        return scaled_frequencies * (1 - smoothing_factor) + inverse_frequencies * smoothing_factor

    @property
    def _attention_scaling_factor(self) -> float:
        return 0.1 * math.log(self.scaling_factor) + 1.0


@dataclass(frozen=True)
class LinearScalingRoPEConfig(RoPEConfigBase):
    scaling_factor: float

    def _scale_inverse_frequencies(
        self,
        inverse_frequencies: Float[Array, " tokens"],
        head_dim: int,  # noqa: ARG002
        max_sequence_length: int,  # noqa: ARG002
    ) -> Float[Array, " tokens"]:
        return inverse_frequencies / self.scaling_factor


RoPEConfig = UnscaledRoPEConfig | LlamaRoPEConfig | YARNRoPEConfig | LinearScalingRoPEConfig

register_config_union(RoPEConfig)



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Literal

from jaxtyping import DTypeLike

from lalamo.modules import (
    DecoderConfig,
    DenseMLPConfig,
    FullPrecisionLinearConfig,
    Identity,
    Mamba2Config,
    MLXQuantizedLinearConfig,
    MLXSemiQuantizedUntiedEmbeddingConfig,
    NormalizationConfig,
    SeparableCausalConvConfig,
    SiLU,
    TiedEmbeddingConfig,
    TransformerConfig,
    TransformerLayerConfig,
    UntiedEmbeddingConfig,
    UpcastMode,
)
from lalamo.quantization import QuantizationMode

from .common import HuggingFaceLMConfig


@dataclass(frozen=True)
class HFLlambaMlpConfig:
    intermediate_size: int
    bias: bool
    act_fn: Literal["silu"]


@dataclass(frozen=True)
class HFLlambaSsmConfig:
    d_state: int
    n_v_heads: int
    n_qk_heads: int
    expand: int
    activation: Literal["identity"]
    bias: bool
    conv_bias: bool = True
    d_conv: int = 4


@dataclass(frozen=True)
class HFLlambaConfig(HuggingFaceLMConfig):
    model_type: Literal["llamba"]
    vocab_size: int
    tie_embeddings: bool
    pad_vocab_size_multiple: int
    lm_head_bias: bool
    d_model: int
    n_layer: int
    resid_dropout: float
    norm_epsilon: float
    mlp_cfg: HFLlambaMlpConfig
    ssm_cfg: HFLlambaSsmConfig

    @property
    def eos_token_ids(self) -> list[int]:
        return [128001, 128008, 128009]

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        metadata_dict: Mapping[str, str],
    ) -> DecoderConfig:
        if "quantization_kwargs.group_size" in metadata_dict:
            embedding_config = MLXSemiQuantizedUntiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                group_size=int(metadata_dict["quantization_kwargs.group_size"]),
                embedding_quantization_mode=QuantizationMode.from_num_bits(
                    int(metadata_dict["quantization_kwargs.bits"]),
                ),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )
        elif self.tie_embeddings:
            embedding_config = TiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                precision=activation_precision,
            )
        else:
            embedding_config = UntiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                precision=activation_precision,
            )

        rmsnorm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.norm_epsilon,
            scale_offset=None,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
            subtract_mean=False,
        )

        if metadata_dict and "quantization_kwargs.group_size" in metadata_dict:
            linear_config = MLXQuantizedLinearConfig(
                group_size=int(metadata_dict["quantization_kwargs.group_size"]),
                weight_quantization_mode=QuantizationMode.from_num_bits(
                    int(metadata_dict["quantization_kwargs.bits"]),
                ),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )
        else:
            linear_config = FullPrecisionLinearConfig(
                precision=activation_precision,
            )

        mlp_config = DenseMLPConfig(
            linear_config=linear_config,
            activation=SiLU(),
            has_up_biases=self.mlp_cfg.bias,
            has_down_biases=self.mlp_cfg.bias,
            up_clipping=None,
            gate_clipping=None,
        )

        inner_dim = self.ssm_cfg.expand * self.d_model
        head_dim = inner_dim // self.ssm_cfg.n_v_heads

        if self.ssm_cfg.activation == "identity":
            activation = Identity()
        elif self.ssm_cfg.activation == "silu":
            activation = SiLU()
        else:
            activation = SiLU()  # fallback

        mamba_config = Mamba2Config(
            in_projection_config=linear_config,
            out_projection_config=linear_config,
            conv_config=SeparableCausalConvConfig(
                precision=activation_precision,
                has_biases=self.ssm_cfg.conv_bias,
            ),
            activation=activation,
            kernel_size=self.ssm_cfg.d_conv,
            num_heads=self.ssm_cfg.n_v_heads,
            num_groups=self.ssm_cfg.n_qk_heads,
            head_dim=head_dim,
            state_dim=self.ssm_cfg.d_state,
            expansion_factor=self.ssm_cfg.expand,
            has_in_biases=self.ssm_cfg.bias,
            has_out_biases=self.ssm_cfg.bias,
        )

        transformer_layer_config = TransformerLayerConfig(
            pre_mixer_norm_config=rmsnorm_config,
            mixer_config=mamba_config,
            post_mixer_norm_config=None,
            pre_mlp_norm_config=rmsnorm_config,
            mlp_config=mlp_config,
            post_mlp_norm_config=None,
        )
        transformer_config = TransformerConfig(
            global_rope_config=None,
            local_rope_config=None,
            layer_configs=(transformer_layer_config,) * self.n_layer,
            output_norm_config=rmsnorm_config,
            model_dim=self.d_model,
            hidden_dim=self.mlp_cfg.intermediate_size,
            context_length=context_length or 4096,
        )

        return DecoderConfig(
            embedding_config=embedding_config,
            transformer_config=transformer_config,
            vocab_size=self.vocab_size,
        )



from abc import abstractmethod

import jax
import jax.numpy as jnp
from attr import dataclass
from jaxtyping import Array, Float

from lalamo.modules.common import register_config_union

__all__ = [
    "GELU",
    "Activation",
    "Identity",
    "SiLU",
]


@dataclass(frozen=True)
class ActivationBase:
    @abstractmethod
    def __call__(self, x: Float[Array, "*dims"]) -> Float[Array, "*dims"]: ...


@dataclass(frozen=True)
class SiLU(ActivationBase):
    alpha: float = 1.0

    def __call__(self, x: Float[Array, "*dims"]) -> Float[Array, "*dims"]:
        return x / (1 + jnp.exp(-x * self.alpha))


@dataclass(frozen=True)
class GELU(ActivationBase):
    def __call__(self, x: Float[Array, "*dims"]) -> Float[Array, "*dims"]:
        return jax.nn.gelu(x)


@dataclass(frozen=True)
class Identity(ActivationBase):
    def __call__(self, x: Float[Array, "*dims"]) -> Float[Array, "*dims"]:
        return x


Activation = SiLU | GELU | Identity


register_config_union(Activation)



from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
import jax
from einops import einsum, rearrange, repeat
from jax import numpy as jnp
from jax import vmap
from jaxtyping import Array, Bool, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import dummy_array
from lalamo.modules.common import ParameterTree, PositionalEmbeddingSelector, require_array, require_tree
from lalamo.modules.linear import LinearBase, LinearConfig
from lalamo.modules.normalization import Normalization, NormalizationConfig
from lalamo.modules.rope import PositionalEmbeddings
from lalamo.modules.utils import apply_soft_capping

from .common import TokenMixerBase, TokenMixerConfigBase, TokenMixerResult
from .state import DynamicKVCacheLayer, KVCacheLayer, StaticKVCacheLayer

__all__ = [
    "Attention",
    "AttentionConfig",
    "AttentionResult",
]


def _repeat_kv(
    keys_or_values: Float[Array, "tokens groups channels"],
    group_size: int,
) -> Float[Array, "tokens groups*group_size channels"]:
    return repeat(
        keys_or_values,
        "tokens groups channels -> tokens (groups group_size) channels",
        group_size=group_size,
    )


def _soft_capped_attention_kernel(
    queries: Float[Array, "dst_tokens heads head_channels"],
    keys: Float[Array, "src_tokens groups head_channels"],
    values: Float[Array, "src_tokens groups head_channels"],
    mask: Bool[Array, "dst_tokens src_tokens"] | None,
    scale: float | None,
    logit_soft_cap: float,
) -> Float[Array, "dst_tokens heads head_channels"]:
    _, num_heads, head_dim = queries.shape
    _, num_groups, _ = keys.shape
    group_size = num_heads // num_groups
    keys = _repeat_kv(keys, group_size)
    values = _repeat_kv(values, group_size)
    queries_head_first = rearrange(queries, "dst_tokens heads channels -> heads dst_tokens channels")
    keys_head_first = rearrange(keys, "src_tokens heads channels -> heads src_tokens channels")
    attention_logits = einsum(
        queries_head_first,
        keys_head_first,
        "heads dst_tokens channels, heads src_tokens channels -> heads dst_tokens src_tokens",
    )
    if mask is not None:
        attention_logits = jnp.where(
            mask,
            attention_logits,
            jnp.array(float("-inf"), dtype=attention_logits.dtype),
        )

    if scale is None:
        scale_val = head_dim**-0.5
    else:
        scale_val = float(scale)
    attention_logits = attention_logits * scale_val
    attention_logits = apply_soft_capping(attention_logits, logit_soft_cap)
    attention_weights = jax.nn.softmax(attention_logits, axis=-1)
    return einsum(
        attention_weights,
        values,
        "heads dst_tokens src_tokens, src_tokens heads channels -> dst_tokens heads channels",
    )


AttentionResult = TokenMixerResult[KVCacheLayer]


@dataclass(frozen=True)
class AttentionConfig(TokenMixerConfigBase):
    qkv_projection_config: LinearConfig
    out_projection_config: LinearConfig

    query_norm_config: NormalizationConfig | None
    key_norm_config: NormalizationConfig | None

    num_heads: int
    num_groups: int
    head_dim: int
    is_causal: bool
    scale: float | None
    sliding_window_size: int | None

    logit_soft_cap: float | None
    has_sinks: bool
    has_qkv_biases: bool
    has_out_biases: bool

    @property
    def rope_dim(self) -> int:
        return self.head_dim

    def random_init(
        self,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "Attention":
        qkv_key, out_key = jax.random.split(key)
        qkv_projection = self.qkv_projection_config.random_init(
            input_dim=model_dim,
            output_dims=(
                self.num_heads * self.head_dim,
                self.num_groups * self.head_dim,
                self.num_groups * self.head_dim,
            ),
            has_biases=self.has_qkv_biases,
            key=qkv_key,
        )
        out_projection = self.out_projection_config.random_init(
            self.num_heads * self.head_dim,
            (model_dim,),
            has_biases=self.has_out_biases,
            key=out_key,
        )

        if self.query_norm_config is not None:
            query_norm = self.query_norm_config.init(
                input_dim=self.head_dim,
            )
        else:
            query_norm = None

        if self.key_norm_config is not None:
            key_norm = self.key_norm_config.init(
                input_dim=self.head_dim,
            )
        else:
            key_norm = None

        if self.has_sinks:
            sinks = jnp.zeros((self.num_heads,), dtype=qkv_projection.activation_precision)
        else:
            sinks = None

        return Attention(
            self,
            qkv_projection=qkv_projection,
            out_projection=out_projection,
            query_norm=query_norm,
            key_norm=key_norm,
            sinks=sinks,
            num_heads=self.num_heads,
            num_groups=self.num_groups,
            head_dim=self.head_dim,
            is_causal=self.is_causal,
            scale=self.scale,
            sliding_window_size=self.sliding_window_size,
        )

    def empty(
        self,
        model_dim: int,
    ) -> "Attention":
        qkv_projection = self.qkv_projection_config.empty(
            input_dim=model_dim,
            output_dims=(
                self.num_heads * self.head_dim,
                self.num_groups * self.head_dim,
                self.num_groups * self.head_dim,
            ),
            has_biases=self.has_qkv_biases,
        )
        out_projection = self.out_projection_config.empty(
            self.num_heads * self.head_dim,
            (model_dim,),
            has_biases=self.has_out_biases,
        )

        if self.query_norm_config is not None:
            query_norm = self.query_norm_config.empty(
                input_dim=self.head_dim,
            )
        else:
            query_norm = None

        if self.key_norm_config is not None:
            key_norm = self.key_norm_config.empty(
                input_dim=self.head_dim,
            )
        else:
            key_norm = None

        if self.has_sinks:
            sinks = dummy_array(self.num_heads, qkv_projection.activation_precision)
        else:
            sinks = None

        return Attention(
            self,
            qkv_projection=qkv_projection,
            out_projection=out_projection,
            query_norm=query_norm,
            key_norm=key_norm,
            sinks=sinks,
            num_heads=self.num_heads,
            num_groups=self.num_groups,
            head_dim=self.head_dim,
            is_causal=self.is_causal,
            scale=self.scale,
            sliding_window_size=self.sliding_window_size,
        )


class Attention(TokenMixerBase[AttentionConfig, KVCacheLayer]):
    qkv_projection: LinearBase
    out_projection: LinearBase

    query_norm: Normalization | None
    key_norm: Normalization | None

    sinks: Float[Array, " heads"] | None

    num_heads: int = eqx.field(static=True)
    num_groups: int = eqx.field(static=True)
    head_dim: int = eqx.field(static=True)

    is_causal: bool = eqx.field(static=True)

    scale: float | None = eqx.field(static=True)
    sliding_window_size: int | None = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.qkv_projection.activation_precision

    @property
    def model_dim(self) -> int:
        return self.qkv_projection.input_dim

    @property
    def group_size(self) -> int:
        return self.num_heads // self.num_groups

    @property
    def use_sliding_window(self) -> bool:
        return self.sliding_window_size is not None

    @property
    def positional_embedding_selector(self) -> PositionalEmbeddingSelector:
        if self.use_sliding_window:
            return PositionalEmbeddingSelector.LOCAL
        return PositionalEmbeddingSelector.GLOBAL

    @property
    def has_sinks(self) -> bool:
        return self.sinks is not None

    def __post_init__(self) -> None:
        if self.qkv_projection.has_biases != self.config.has_qkv_biases:
            raise ValueError(
                f"QKV projection has_biases {self.qkv_projection.has_biases} does not match"
                f" the specified config has_qkv_biases {self.config.has_qkv_biases}",
            )
        if self.out_projection.has_biases != self.config.has_out_biases:
            raise ValueError(
                f"Output projection has_biases {self.out_projection.has_biases} does not match"
                f" the specified config has_out_biases {self.config.has_out_biases}",
            )
        if self.query_norm is not None and self.query_norm.input_dim != self.head_dim:
            raise ValueError(
                f"Query normalization input dimension must match head_dim ({self.head_dim}),"
                f" got {self.query_norm.input_dim}",
            )
        if self.key_norm is not None and self.key_norm.input_dim != self.head_dim:
            raise ValueError(
                f"Key normalization input dimension must match head_dim ({self.head_dim}),"
                f" got {self.key_norm.input_dim}",
            )
        if self.num_heads % self.num_groups != 0:
            raise ValueError(
                "Number of heads must be divisible by the number of groups,"
                f" got {self.num_heads} heads and {self.num_groups} groups",
            )
        if self.out_projection.input_dim != self.num_heads * self.head_dim:
            raise ValueError(
                f"Output projection input dimension must be num_heads * head_dim"
                f" ({self.num_heads} * {self.head_dim} = {self.num_heads * self.head_dim}),"
                f" got {self.out_projection.input_dim}",
            )
        q_output_dim, k_output_dim, v_output_dim = self.qkv_projection.output_dims
        if q_output_dim != self.num_heads * self.head_dim:
            raise ValueError(
                f"Query projection output dimension must be num_heads * head_dim"
                f" ({self.num_heads} * {self.head_dim} = {self.num_heads * self.head_dim}),"
                f" got {q_output_dim}",
            )
        if k_output_dim != self.num_groups * self.head_dim:
            raise ValueError(
                f"Key projection output dimension must be num_groups * head_dim"
                f" ({self.num_groups} * {self.head_dim} = {self.num_groups * self.head_dim}),"
                f" got {k_output_dim}",
            )
        if v_output_dim != self.num_groups * self.head_dim:
            raise ValueError(
                f"Value projection output dimension must be num_groups * head_dim"
                f" ({self.num_groups} * {self.head_dim} = {self.num_groups * self.head_dim}),"
                f" got {v_output_dim}",
            )
        if self.sinks is not None:
            (num_sink_heads,) = self.sinks.shape
            if num_sink_heads != self.num_heads:
                raise ValueError(
                    f"Number of sink heads must be equal to number of heads ({self.num_heads}), got {num_sink_heads}",
                )

    @eqx.filter_jit
    def __call__(
        self,
        inputs: Float[Array, "suffix_tokens channels"],
        positional_embeddings: PositionalEmbeddings | None,
        state: KVCacheLayer | None = None,
        return_updated_state: bool = False,
        length_without_padding: Int[Array, ""] | int | None = None,
    ) -> AttentionResult:
        queries, keys, values = vmap(self.qkv_projection, in_axes=0)(inputs)
        queries = rearrange(
            queries,
            "tokens (heads head_channels) -> tokens heads head_channels",
            heads=self.num_heads,
            head_channels=self.head_dim,
        )
        keys = rearrange(
            keys,
            "tokens (groups head_channels) -> tokens groups head_channels",
            groups=self.num_groups,
            head_channels=self.head_dim,
        )
        values = rearrange(
            values,
            "tokens (groups head_channels) -> tokens groups head_channels",
            groups=self.num_groups,
            head_channels=self.head_dim,
        )

        if self.query_norm is not None:
            queries = vmap(vmap(self.query_norm))(queries)
        if self.key_norm is not None:
            keys = vmap(vmap(self.key_norm))(keys)

        if positional_embeddings is not None:
            apply_positional_embeddings = vmap(positional_embeddings.apply, in_axes=1, out_axes=1)
            queries = apply_positional_embeddings(queries)
            keys = apply_positional_embeddings(keys)

        if state is None:
            updated_state = DynamicKVCacheLayer.init(self.has_sinks, keys, values, length=length_without_padding)
        else:
            updated_state = state.extend(keys, values, added_length=length_without_padding)

        num_suffix_tokens, _, _ = queries.shape
        mask = updated_state.attention_mask(
            num_suffix_tokens,
            self.is_causal,
            length_without_padding,
            self.sliding_window_size,
        )
        if self.sinks is not None:
            sink_bias = jnp.zeros((self.num_heads, *mask.shape), dtype=queries.dtype)
            sink_bias = sink_bias.at[:, :, 0].set(self.sinks[:, None])
        else:
            sink_bias = None

        if self.config.logit_soft_cap is not None:
            attention_output = _soft_capped_attention_kernel(
                queries,
                updated_state.keys,
                updated_state.values,
                mask=mask,
                scale=self.scale,
                logit_soft_cap=self.config.logit_soft_cap,
            )
        else:
            attention_output = jax.nn.dot_product_attention(
                queries,
                updated_state.keys,
                updated_state.values,
                bias=sink_bias,
                mask=mask,
                scale=self.scale,
            )
        attention_output = rearrange(
            attention_output,
            "tokens heads head_channels -> tokens (heads head_channels)",
            heads=self.num_heads,
            head_channels=self.head_dim,
        )
        (result,) = vmap(self.out_projection, in_axes=0)(attention_output)

        if not return_updated_state:
            updated_state = None

        return AttentionResult(
            outputs=result,
            state=updated_state,
        )

    def init_static_state(self, capacity: int) -> StaticKVCacheLayer:
        return StaticKVCacheLayer.init(
            self.has_sinks,
            capacity,
            self.num_groups,
            self.head_dim,
            self.activation_precision,
        )

    def export_weights(self) -> ParameterTree:
        result: dict[str, ParameterTree | Array] = {
            "qkv_projection": self.qkv_projection.export_weights(),
            "out_projection": self.out_projection.export_weights(),
        }
        if self.query_norm is not None:
            result["query_norm"] = self.query_norm.export_weights()
        if self.key_norm is not None:
            result["key_norm"] = self.key_norm.export_weights()
        if self.sinks is not None:
            assert isinstance(self.sinks, Array)
            result["sinks"] = self.sinks
        return result

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            qkv_projection=self.qkv_projection.import_weights(require_tree(weights["qkv_projection"])),
            out_projection=self.out_projection.import_weights(require_tree(weights["out_projection"])),
            query_norm=self.query_norm.import_weights(require_tree(weights["query_norm"]))
            if self.query_norm
            else None,
            key_norm=self.key_norm.import_weights(require_tree(weights["key_norm"])) if self.key_norm else None,
            sinks=require_array(weights["sinks"]) if self.sinks is not None else None,
        )



from collections.abc import Iterable, Iterator, Mapping
from dataclasses import dataclass, replace

import jax.numpy as jnp
from einops import rearrange
from jaxtyping import Array, Float, Int

from lalamo.common import ParameterPath
from lalamo.modules import (
    Attention,
    Decoder,
    DenseMLP,
    Normalization,
    QLoRALinear,
    QuantizedTiedEmbedding,
    TransformerLayer,
)

from .common import load_parameters

__all__ = ["load_executorch"]


@dataclass(frozen=True)
class QLoRALinearParams:
    weights: Int[Array, "out_channels in_channels"]
    zero_points: Int[Array, "out_channels groups"]
    scales: Float[Array, "out_channels groups"]
    lora_down_weights: Float[Array, "total_lora_channels in_channels"]
    lora_up_weights: tuple[Float[Array, "..."], ...]

    def __iter__(self) -> Iterator[Array]:
        yield self.weights
        yield self.zero_points
        yield self.scales
        yield self.lora_down_weights
        yield from self.lora_up_weights

    def __len__(self) -> int:
        return 3 + len(self.lora_up_weights)


def params_selector(module: QLoRALinear) -> tuple:
    return (
        module.weights,
        module.zero_points,
        module.scales,
        module.lora_down_weights,
        *module.lora_up_weights,
    )


def get_qlora_linear_params(
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    weights_dtype: jnp.dtype,
) -> QLoRALinearParams:
    shift_to_unsigned = 8

    weights = weights_dict[path / "weight"].astype(weights_dtype)
    scales = weights_dict[path / "scales"]

    # We don't support signed int4 on the inference side, so we map int4 to uint4 and add zero-points.
    weights = weights + shift_to_unsigned
    zero_points = jnp.ones_like(scales) * shift_to_unsigned

    lora_down_weights = weights_dict[path / "adaptor" / "A" / "weight"]
    lora_up_weights = (weights_dict[path / "adaptor" / "B" / "weight"],)
    return QLoRALinearParams(weights, scales, zero_points, lora_down_weights, lora_up_weights)


def merge_linear_params(params_list: Iterable[QLoRALinearParams]) -> QLoRALinearParams:
    params_list = list(params_list)
    weights = jnp.concatenate([p.weights for p in params_list], axis=0)
    scales = jnp.concatenate([p.scales for p in params_list], axis=0)
    zero_points = jnp.concatenate([p.zero_points for p in params_list], axis=0)
    lora_down_weights = jnp.concatenate([p.lora_down_weights for p in params_list], axis=0)
    lora_up_weights = tuple(w for p in params_list for w in p.lora_up_weights)
    return QLoRALinearParams(weights, scales, zero_points, lora_down_weights, lora_up_weights)


def load_linear(module: QLoRALinear, weights_dict: dict[str, Array], path: ParameterPath) -> QLoRALinear:
    params = get_qlora_linear_params(weights_dict, path, module.weights.dtype)
    return load_parameters(params_selector, module, params)


def load_mlp(module: DenseMLP, weights_dict: Mapping[str, Array], path: ParameterPath) -> DenseMLP:
    if not isinstance(module.up_projection, QLoRALinear):
        raise TypeError(f"Expected up_projection to be QLoRALinear, got {type(module.up_projection)}")
    if not isinstance(module.down_projection, QLoRALinear):
        raise TypeError(f"Expected down_projection to be QLoRALinear, got {type(module.down_projection)}")

    up_proj_params = get_qlora_linear_params(weights_dict, path / "w3", module.up_projection.weights.dtype)
    gate_proj_params = get_qlora_linear_params(weights_dict, path / "w1", module.down_projection.weights.dtype)
    down_proj_params = get_qlora_linear_params(weights_dict, path / "w2", module.down_projection.weights.dtype)

    fused_up_gate_params = merge_linear_params([up_proj_params, gate_proj_params])

    return load_parameters(
        lambda m: (*params_selector(m.up_projection), *params_selector(m.down_projection)),
        module,
        (*fused_up_gate_params, *down_proj_params),
    )


def load_rmsnorm(module: Normalization, weights_dict: Mapping[str, Array], path: ParameterPath) -> Normalization:
    return load_parameters(lambda m: (m.scales,), module, (weights_dict[path / "weight"],))


def permute_qk_weights(weights: Array, input_dim: int, num_heads: int, head_dim: int) -> Array:
    # Reference: https://github.com/huggingface/transformers/blob/15bd3e61f8d3680ca472c9314ad07584d20f7b81/src/transformers/models/llama/convert_llama_weights_to_hf.py#L222
    return rearrange(
        weights,
        "(heads rotors reim) input_channels -> (heads reim rotors) input_channels",
        heads=num_heads,
        rotors=head_dim // 2,
        reim=2,
        input_channels=input_dim,
    )


def permute_qk_params(
    *,
    params: QLoRALinearParams,
    model_dim: int,
    num_heads: int,
    head_dim: int,
    quantization_group_size: int,
    lora_rank: int,
) -> QLoRALinearParams:
    # Read https://github.com/huggingface/transformers/issues/25199 to understand WTF is going on here
    return replace(
        params,
        weights=permute_qk_weights(params.weights, model_dim, num_heads, head_dim),
        scales=permute_qk_weights(params.scales, model_dim // quantization_group_size, num_heads, head_dim),
        lora_up_weights=tuple(permute_qk_weights(w, lora_rank, num_heads, head_dim) for w in params.lora_up_weights),
    )


def load_attention(
    module: Attention,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> Attention:
    if not isinstance(module.qkv_projection, QLoRALinear):
        raise TypeError(f"Expected qkv_projection to be QLoRALinear, got {type(module.qkv_projection)}")

    model_dim = module.model_dim
    num_heads = module.num_heads
    num_groups = module.num_groups
    head_dim = module.head_dim
    lora_rank = module.qkv_projection.config.lora_rank

    q_params = get_qlora_linear_params(weights_dict, path / "wq", module.qkv_projection.weights.dtype)
    q_params = permute_qk_params(
        params=q_params,
        model_dim=model_dim,
        num_heads=num_heads,
        head_dim=head_dim,
        quantization_group_size=module.qkv_projection.config.group_size,
        lora_rank=lora_rank,
    )

    k_params = get_qlora_linear_params(weights_dict, path / "wk", module.qkv_projection.weights.dtype)
    k_params = permute_qk_params(
        params=k_params,
        model_dim=model_dim,
        num_heads=num_groups,
        head_dim=head_dim,
        quantization_group_size=module.qkv_projection.config.group_size,
        lora_rank=lora_rank,
    )

    v_params = get_qlora_linear_params(weights_dict, path / "wv", module.qkv_projection.weights.dtype)

    out_params = get_qlora_linear_params(weights_dict, path / "wo", module.qkv_projection.weights.dtype)

    qkv_params = merge_linear_params([q_params, k_params, v_params])
    return load_parameters(
        lambda m: (*params_selector(m.qkv_projection), *params_selector(m.out_projection)),
        module,
        (*qkv_params, *out_params),
    )


def load_transformer_layer(
    module: TransformerLayer,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> TransformerLayer:
    if module.post_mixer_norm is not None:
        raise ValueError("Post attention normalization is not supported")
    if module.post_mlp_norm is not None:
        raise ValueError("Post MLP normalization is not supported")
    if module.pre_mixer_norm is not None:
        attention_norm = load_rmsnorm(module.pre_mixer_norm, weights_dict, path / "attention_norm")
    else:
        attention_norm = None
    assert isinstance(module.mixer, Attention)
    attention = load_attention(module.mixer, weights_dict, path / "mixer")
    mlp_norm = load_rmsnorm(module.pre_mlp_norm, weights_dict, path / "ffn_norm")
    assert isinstance(module.mlp, DenseMLP)
    mlp = load_mlp(module.mlp, weights_dict, path / "feed_forward")
    return load_parameters(
        lambda m: (m.pre_mixer_norm, m.mixer, m.pre_mlp_norm, m.mlp),
        module,
        (attention_norm, attention, mlp_norm, mlp),
    )


def load_embedding(
    module: QuantizedTiedEmbedding,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> QuantizedTiedEmbedding:
    weights = weights_dict[path / "weight"].astype(module.weights.dtype)
    scales = weights_dict[path / "scales"].squeeze(1)

    return load_parameters(lambda m: (m.weights, m.scales), module, (weights, scales))


def load_executorch(module: Decoder, weights_dict: Mapping[str, Array]) -> Decoder:
    root_path = ParameterPath()
    if not isinstance(module.embedding, QuantizedTiedEmbedding):
        raise TypeError(f"Expected embedding to be QuantizedTiedEmbedding, got {type(module.embedding)}")

    embedding = load_embedding(module.embedding, weights_dict, root_path / "tok_embeddings")
    transformer_layers = tuple(
        load_transformer_layer(layer, weights_dict, root_path / f"layers.{i}")
        for i, layer in enumerate(module.transformer.layers)
    )
    output_norm = load_rmsnorm(module.transformer.output_norm, weights_dict, root_path / "norm")
    return load_parameters(
        lambda m: (m.embedding, m.transformer.layers, m.transformer.output_norm),
        module,
        (embedding, transformer_layers, output_norm),
    )



import math
from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass, replace
from functools import partial
from typing import Self

import equinox as eqx
import jax
import jax.numpy as jnp
from einops import rearrange
from jax import vmap
from jaxtyping import Array, Bool, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree, require_tree
from lalamo.modules.utils import vmap_twice

from .activations import Activation
from .common import (
    DummyUnionMember,
    ForwardPassMode,
    LalamoModule,
    register_config_union,
)
from .linear import LinearBase, LinearConfig

__all__ = [
    "DenseMLP",
    "DenseMLPConfig",
    "MLPBase",
    "MLPConfig",
    "MLPForwardPassConfig",
    "MixtureOfExperts",
    "MixtureOfExpertsConfig",
    "RoutingFunction",
    "SoftmaxRouting",
]


_SENTINEL = 2**31 - 1


@dataclass(frozen=True)
class MLPForwardPassConfig:
    moe_chunk_size_ratio: float = 0.2


class MLPBase[ConfigT: MLPConfig](LalamoModule[ConfigT]):
    @property
    @abstractmethod
    def activation_precision(self) -> DTypeLike: ...

    @property
    @abstractmethod
    def model_dim(self) -> int: ...

    @property
    @abstractmethod
    def hidden_dim(self) -> int: ...

    @abstractmethod
    def __call__(
        self,
        inputs: Float[Array, "batch suffix_tokens channels"],
        lengths_without_padding: Int[Array, " batch"] | None = None,
        forward_pass_mode: ForwardPassMode = ForwardPassMode.MULTI_TOKEN,
        forward_pass_config: MLPForwardPassConfig | None = None,
    ) -> Float[Array, "batch suffix_tokens channels"]: ...


@dataclass(frozen=True)
class MLPConfigBase(ABC):
    @abstractmethod
    def random_init(self, model_dim: int, hidden_dim: int, *, key: PRNGKeyArray) -> MLPBase: ...

    @abstractmethod
    def empty(self, model_dim: int, hidden_dim: int) -> MLPBase: ...


@dataclass(frozen=True)
class DenseMLPConfig(MLPConfigBase):
    linear_config: LinearConfig
    activation: Activation
    has_up_biases: bool
    has_down_biases: bool
    gate_clipping: tuple[float | None, float | None] | None
    up_clipping: tuple[float | None, float | None] | None

    def random_init(self, model_dim: int, hidden_dim: int, *, key: PRNGKeyArray) -> "DenseMLP":
        up_key, down_key = jax.random.split(key)
        return DenseMLP(
            self,
            up_projection=self.linear_config.random_init(
                model_dim,
                (hidden_dim, hidden_dim),
                has_biases=self.has_up_biases,
                key=up_key,
            ),
            down_projection=self.linear_config.random_init(
                hidden_dim,
                (model_dim,),
                has_biases=self.has_down_biases,
                key=down_key,
            ),
        )

    def empty(self, model_dim: int, hidden_dim: int) -> "DenseMLP":
        return DenseMLP(
            self,
            up_projection=self.linear_config.empty(
                model_dim,
                (hidden_dim, hidden_dim),
                has_biases=self.has_up_biases,
            ),
            down_projection=self.linear_config.empty(
                hidden_dim,
                (model_dim,),
                has_biases=self.has_down_biases,
            ),
        )

    def random_init_mixture(
        self,
        mixture_size: int,
        model_dim: int,
        hidden_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "DenseMLP":
        up_key, down_key = jax.random.split(key)
        return DenseMLP(
            self,
            up_projection=self.linear_config.random_init_mixture(
                mixture_size,
                model_dim,
                (hidden_dim, hidden_dim),
                has_biases=self.has_up_biases,
                key=up_key,
            ),
            down_projection=self.linear_config.random_init_mixture(
                mixture_size,
                hidden_dim,
                (model_dim,),
                has_biases=self.has_down_biases,
                key=down_key,
            ),
        )

    def empty_mixture(
        self,
        mixture_size: int,
        model_dim: int,
        hidden_dim: int,
    ) -> "DenseMLP":
        return DenseMLP(
            self,
            up_projection=self.linear_config.empty_mixture(
                mixture_size,
                model_dim,
                (hidden_dim, hidden_dim),
                has_biases=self.has_up_biases,
            ),
            down_projection=self.linear_config.empty_mixture(
                mixture_size,
                hidden_dim,
                (model_dim,),
                has_biases=self.has_down_biases,
            ),
        )


class DenseMLP(MLPBase[DenseMLPConfig]):
    up_projection: LinearBase
    down_projection: LinearBase

    @property
    def activation_precision(self) -> DTypeLike:
        return self.up_projection.activation_precision

    @property
    def model_dim(self) -> int:
        return self.up_projection.input_dim

    @property
    def hidden_dim(self) -> int:
        return self.down_projection.input_dim

    @property
    def mixture_size(self) -> int | None:
        return self.up_projection.mixture_size

    def __post_init__(self) -> None:
        up_output_dim, gate_output_dim = self.up_projection.output_dims
        if up_output_dim != gate_output_dim:
            raise ValueError(
                f"Up projection output dimension {up_output_dim} does not match"
                f" the gate output dimension {gate_output_dim}",
            )
        (down_output_dim,) = self.down_projection.output_dims
        if (self.up_projection.input_dim, up_output_dim) != (
            down_output_dim,
            self.down_projection.input_dim,
        ):
            raise ValueError(
                f"Down projection dimensions {self.down_projection.input_dim, down_output_dim} do not match"
                f" the up projection output dimensions {self.up_projection.input_dim, up_output_dim}",
            )

    @eqx.filter_jit
    def __call__(
        self,
        inputs: Float[Array, "batch suffix_tokens channels"],
        lengths_without_padding: Int[Array, " batch"] | None = None,  # noqa: ARG002
        forward_pass_mode: ForwardPassMode = ForwardPassMode.MULTI_TOKEN,  # noqa: ARG002
        forward_pass_config: MLPForwardPassConfig | None = None,  # noqa: ARG002
    ) -> Float[Array, "batch suffix_tokens channels"]:
        return vmap_twice(self.call_unbatched)(inputs)

    @eqx.filter_jit
    def call_unbatched(
        self,
        inputs: Float[Array, " channels"],
    ) -> Float[Array, " channels"]:
        if self.mixture_size is not None:
            raise ValueError(
                "Mixtures of linear layers cannot be called directly."
                "They are intended to be used with methods eqx.filter_vmap or lax.scan instead.",
            )
        up_proj, gate = self.up_projection(inputs)
        if self.config.gate_clipping:
            gate = jnp.clip(gate, *self.config.gate_clipping)
        if self.config.up_clipping:
            up_proj = jnp.clip(up_proj, *self.config.up_clipping)
        gate = self.config.activation(gate)
        (result,) = self.down_projection(up_proj * gate)

        return result

    def export_weights(self) -> ParameterTree:
        return {
            "up_projection": self.up_projection.export_weights(),
            "down_projection": self.down_projection.export_weights(),
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            up_projection=self.up_projection.import_weights(require_tree(weights["up_projection"])),
            down_projection=self.down_projection.import_weights(require_tree(weights["down_projection"])),
        )


class RoutingMap(eqx.Module):
    expert_mask: Bool[Array, "*batch_tokens experts"]
    expert_weights: Float[Array, "*batch_tokens experts"]


@dataclass(frozen=True)
class RoutingFunctionBase(ABC):
    def __call__(self, logits: Float[Array, "batch_tokens experts"], num_active: int) -> RoutingMap:
        return vmap(partial(self.call_unbatched, num_active=num_active))(logits)

    @abstractmethod
    def call_unbatched(self, logits: Float[Array, " experts"], num_active: int) -> RoutingMap: ...


@dataclass(frozen=True)
class SoftmaxRouting(RoutingFunctionBase):
    def call_unbatched(self, logits: Float[Array, " experts"], num_active: int) -> RoutingMap:
        active_logits, active_indices = jax.lax.top_k(logits, num_active)
        active_weights = jax.nn.softmax(active_logits)
        mask = jnp.zeros_like(logits, dtype=bool)
        mask = mask.at[active_indices].set(True)
        expert_weights = jnp.zeros_like(logits)
        expert_weights = expert_weights.at[active_indices].set(active_weights)
        return RoutingMap(expert_mask=mask, expert_weights=expert_weights)


RoutingFunction = SoftmaxRouting | DummyUnionMember


register_config_union(RoutingFunction)


@dataclass(frozen=True)
class MixtureOfExpertsConfig(ABC):
    expert_config: DenseMLPConfig
    router_config: LinearConfig
    routing_function: RoutingFunction

    mixture_size: int
    num_experts_per_token: int
    router_has_biases: bool

    def random_init(self, model_dim: int, hidden_dim: int, *, key: PRNGKeyArray) -> "MixtureOfExperts":
        experts_key, router_key = jax.random.split(key)
        router = self.router_config.random_init(
            model_dim,
            (self.mixture_size,),
            has_biases=self.router_has_biases,
            key=router_key,
        )
        experts = self.expert_config.random_init_mixture(self.mixture_size, model_dim, hidden_dim, key=experts_key)
        return MixtureOfExperts(self, router, experts)

    def empty(self, model_dim: int, hidden_dim: int) -> "MixtureOfExperts":
        router = self.router_config.empty(model_dim, (self.mixture_size,), has_biases=self.router_has_biases)
        experts = self.expert_config.empty_mixture(self.mixture_size, model_dim, hidden_dim)
        return MixtureOfExperts(self, router, experts)


class MixtureOfExperts(MLPBase[MixtureOfExpertsConfig]):
    router: LinearBase
    experts: DenseMLP

    @property
    def mixture_size(self) -> int:
        return self.config.mixture_size

    @property
    def num_experts_per_token(self) -> int:
        return self.config.num_experts_per_token

    @property
    def activation_precision(self) -> DTypeLike:
        return self.experts.activation_precision

    @property
    def model_dim(self) -> int:
        return self.experts.model_dim

    @property
    def hidden_dim(self) -> int:
        return self.experts.hidden_dim

    def __post_init__(self) -> None:
        if self.router.input_dim != self.experts.model_dim:
            raise ValueError(
                f"Router input dimension ({self.router.input_dim}) must match experts model_dim"
                f" ({self.experts.model_dim}).",
            )

        (router_output_dim,) = self.router.output_dims
        if router_output_dim != self.mixture_size:
            raise ValueError(
                f"Router output dimension ({router_output_dim}) must equal mixture_size ({self.mixture_size}).",
            )

        if self.experts.mixture_size != self.mixture_size:
            raise ValueError(
                f"Experts mixture_size ({self.experts.mixture_size}) does not match specified mixture_size"
                f" ({self.mixture_size}).",
            )

    def __call__(
        self,
        inputs: Float[Array, "batch suffix_tokens channels"],
        lengths_without_padding: Int[Array, " batch"] | None = None,
        forward_pass_mode: ForwardPassMode = ForwardPassMode.MULTI_TOKEN,
        forward_pass_config: MLPForwardPassConfig | None = None,
    ) -> Float[Array, "batch suffix_tokens channels"]:
        match forward_pass_mode:
            case ForwardPassMode.MULTI_TOKEN:
                return self.call_prefill_mode(inputs, lengths_without_padding, forward_pass_config)
            case ForwardPassMode.SINGLE_TOKEN:
                return self.call_decode_mode(inputs)

    @eqx.filter_jit
    def call_decode_mode(
        self,
        inputs: Float[Array, "batch suffix_tokens channels"],
    ) -> Float[Array, "batch suffix_tokens channels"]:
        def per_token(x: Float[Array, " channels"]) -> Float[Array, " channels"]:
            (router_logits,) = self.router(x)
            routing = self.config.routing_function.call_unbatched(
                router_logits,
                num_active=self.num_experts_per_token,
            )
            active_indices = jnp.flatnonzero(routing.expert_mask, size=self.num_experts_per_token)
            active_weights = routing.expert_weights[active_indices]

            def apply_one(idx: Int[Array, ""], w: Float[Array, ""]) -> Float[Array, " channels"]:
                selected_expert = jax.tree_util.tree_map(
                    lambda leaf: jax.lax.dynamic_index_in_dim(leaf, idx, axis=0, keepdims=False),
                    self.experts,
                )
                return selected_expert.call_unbatched(x) * w

            contributions = vmap(apply_one)(active_indices, active_weights)
            return jnp.sum(contributions, axis=0)

        return vmap_twice(per_token)(inputs)

    @eqx.filter_jit
    def call_prefill_mode(
        self,
        inputs: Float[Array, "batch suffix_tokens channels"],
        lengths_without_padding: Int[Array, " batch"] | None = None,
        forward_pass_config: MLPForwardPassConfig | None = None,
    ) -> Float[Array, "batch suffix_tokens channels"]:
        forward_pass_config = forward_pass_config or MLPForwardPassConfig()
        batch_size, sequence_length, _ = inputs.shape
        num_tokens = batch_size * sequence_length
        if lengths_without_padding is None:
            lengths_without_padding = jnp.ones(batch_size, dtype=jnp.int32) * sequence_length
        padding_mask = jnp.arange(sequence_length)[None, :] < lengths_without_padding[:, None]

        flattened_inputs = rearrange(inputs, "batch suffix_tokens channels -> (batch suffix_tokens) channels")
        flattened_padding_mask = rearrange(padding_mask, "batch suffix_tokens -> (batch suffix_tokens)")

        (router_logits,) = vmap(self.router)(flattened_inputs)
        routing_map = self.config.routing_function(router_logits, self.num_experts_per_token)
        token_mask = rearrange(
            routing_map.expert_mask & flattened_padding_mask[:, None],
            "tokens experts -> experts tokens",
        )
        expert_weights = rearrange(
            routing_map.expert_weights,
            "tokens experts -> experts tokens",
        )
        expert_weights = jnp.where(token_mask, expert_weights, 0.0)

        chunk_size = math.ceil(num_tokens * forward_pass_config.moe_chunk_size_ratio)
        num_padded_tokens = math.ceil(num_tokens / chunk_size) * chunk_size
        token_indices = vmap(lambda m: jnp.flatnonzero(m, size=num_padded_tokens, fill_value=_SENTINEL))(token_mask)
        chunked_token_indices = rearrange(
            token_indices,
            "experts (chunks chunk_tokens) -> chunks experts chunk_tokens",
            chunk_tokens=chunk_size,
        )

        def loop_iteration(
            accumulator: Float[Array, "tokens channels"],
            token_indices_for_chunk: Int[Array, "experts chunk_tokens"],
        ) -> tuple[Float[Array, "tokens channels"], None]:
            def inner() -> Float[Array, "tokens channels"]:
                weights_for_chunk = jnp.take_along_axis(
                    expert_weights,
                    token_indices_for_chunk,
                    axis=1,
                    mode="fill",
                    fill_value=0.0,
                )

                def run_expert(
                    expert: DenseMLP,
                    indices: Int[Array, " tokens_per_chunk"],
                    weights: Float[Array, " tokens_per_chunk"],
                ) -> Float[Array, "tokens_per_chunk channels"]:
                    inputs = flattened_inputs.at[indices].get(mode="fill", fill_value=0.0)
                    return vmap(expert.call_unbatched)(inputs) * weights[:, None]

                expert_outputs = vmap(run_expert)(self.experts, token_indices_for_chunk, weights_for_chunk)
                return accumulator.at[token_indices_for_chunk].add(
                    expert_outputs,
                    mode="drop",
                )

            return (
                jax.lax.cond(
                    jnp.any(token_indices_for_chunk != _SENTINEL),
                    inner,
                    lambda: accumulator,
                ),
                None,
            )

        result, _ = jax.lax.scan(loop_iteration, jnp.zeros_like(flattened_inputs), chunked_token_indices)
        return rearrange(
            result,
            "(batch suffix_tokens) channels -> batch suffix_tokens channels",
            batch=batch_size,
        )

    def export_weights(
        self,
    ) -> ParameterTree[Array]:
        return {
            "router": self.router.export_weights(),
            "experts": self.experts.export_weights(),
        }

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            router=self.router.import_weights(require_tree(weights["router"])),
            experts=self.experts.import_weights(require_tree(weights["experts"])),
        )


MLPConfig = DenseMLPConfig | MixtureOfExpertsConfig


register_config_union(MLPConfig)



from dataclasses import dataclass
from typing import Any, ClassVar

import cattrs
import requests


@dataclass(frozen=True)
class RegistryModelFile:
    name: str
    url: str
    size: int
    crc32c: str


@dataclass(frozen=True)
class RegistryModel:
    _converter: ClassVar[cattrs.Converter] = cattrs.Converter()

    id: str
    vendor: str
    name: str
    family: str
    size: str
    repo_id: str
    quantization: str | None
    files: list[RegistryModelFile]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "RegistryModel":
        if "repoId" in data:
            data = {**data, "repo_id": data.pop("repoId")}
        return cls._converter.structure(data, cls)


def fetch_available_models() -> list[RegistryModel]:
    api_url = "https://sdk.trymirai.com/api/v1/models/list/lalamo"
    response = requests.get(api_url, timeout=30)
    response.raise_for_status()

    data = response.json()
    models_data = data.get("models", [])

    return [RegistryModel.from_dict(model_data) for model_data in models_data]



import os

# Must run before importing jax / tensorflow, this hides the XLA optimization logs
os.environ.setdefault("TF_CPP_MIN_LOG_LEVEL", "3")

from lalamo.commands import (
    CollectTracesCallbacks,
    ConversionCallbacks,
    EstimateBatchsizeCallbacks,
    Precision,
    TrainCallbacks,
    collect_traces,
    convert,
    estimate_batchsize,
    train,
)
from lalamo.message_processor import (
    AssistantMessage,
    ContentBlock,
    Image,
    Message,
    SystemMessage,
    ToolSchema,
    UserMessage,
)
from lalamo.model_import import ModelSpec, import_model
from lalamo.model_import.model_specs.common import ConfigMap, FileSpec, JSONFieldSpec, ModelType, UseCase, WeightsType
from lalamo.models import ClassifierModel, LanguageModel
from lalamo.quantization import QuantizationMode
from lalamo.speculator import (
    CollectTracesEvent,
    SpeculatorTrainingEvent,
)

__version__ = "0.6.6"

__all__ = [
    "AssistantMessage",
    "ClassifierModel",
    "CollectTracesCallbacks",
    "CollectTracesEvent",
    "ConfigMap",
    "ContentBlock",
    "ConversionCallbacks",
    "EstimateBatchsizeCallbacks",
    "FileSpec",
    "Image",
    "JSONFieldSpec",
    "LanguageModel",
    "Message",
    "ModelSpec",
    "ModelType",
    "Precision",
    "QuantizationMode",
    "SpeculatorTrainingEvent",
    "SystemMessage",
    "ToolSchema",
    "TrainCallbacks",
    "UseCase",
    "UserMessage",
    "WeightsType",
    "collect_traces",
    "convert",
    "estimate_batchsize",
    "import_model",
    "train",
]



from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
import jax
from jax import vmap
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree, require_tree

from .common import ForwardPassMode, LalamoModule
from .embedding import EmbeddingBase, EmbeddingConfig
from .rope import PositionalEmbeddings
from .token_mixers import State
from .transformer import (
    Transformer,
    TransformerConfig,
    TransformerForwardPassConfig,
    TransformerLayerResult,
)
from .utils import vmap_twice

__all__ = [
    "Decoder",
    "DecoderActivationTrace",
    "DecoderConfig",
    "DecoderForwardPassConfig",
    "DecoderResult",
]


type DecoderForwardPassConfig = TransformerForwardPassConfig


class DecoderActivationTrace(eqx.Module):
    token_ids: Int[Array, "batch suffix_tokens"]
    token_positions: Int[Array, "batch suffix_tokens"]
    state: State | None

    local_positional_embeddings: PositionalEmbeddings | None
    global_positional_embeddings: PositionalEmbeddings | None

    layer_results: tuple[TransformerLayerResult, ...]

    output_norm: Float[Array, "batch suffix_tokens channels"]

    def export(self) -> ParameterTree:
        result: dict[str, ParameterTree | Array] = dict(
            token_ids=self.token_ids,
            token_positions=self.token_positions,
            layer_results=[layer_result.export() for layer_result in self.layer_results],
            output_norm=self.output_norm,
        )
        if self.state is not None:
            result["state"] = [state_layer.export() for state_layer in self.state]
        if self.local_positional_embeddings is not None:
            result["local_positional_embeddings"] = self.local_positional_embeddings.export()
        if self.global_positional_embeddings is not None:
            result["global_positional_embeddings"] = self.global_positional_embeddings.export()
        return result


class DecoderResult(eqx.Module):
    logits: Float[Array, "batch suffix_tokens channels"]
    updated_state: State | None = None
    activation_trace: DecoderActivationTrace | None = None

    def export(self) -> ParameterTree:
        result: dict[str, ParameterTree | Array] = dict(
            logits=self.logits,
        )
        if self.updated_state is not None:
            result["updated_state"] = [state_layer.export() for state_layer in self.updated_state]
        if self.activation_trace is not None:
            result["activation_trace"] = self.activation_trace.export()
        return result


@dataclass(frozen=True)
class DecoderConfig:
    embedding_config: EmbeddingConfig
    transformer_config: TransformerConfig

    vocab_size: int

    def random_init(
        self,
        *,
        key: PRNGKeyArray,
    ) -> "Decoder":
        embedding_key, transformer_key = jax.random.split(key)
        embedding = self.embedding_config.random_init(
            vocab_size=self.vocab_size,
            model_dim=self.transformer_config.model_dim,
            key=embedding_key,
        )
        transformer = self.transformer_config.random_init(key=transformer_key)

        return Decoder(
            config=self,
            embedding=embedding,
            transformer=transformer,
        )

    def empty(self) -> "Decoder":
        embedding = self.embedding_config.empty(
            vocab_size=self.vocab_size,
            model_dim=self.transformer_config.model_dim,
        )
        transformer = self.transformer_config.empty()

        return Decoder(
            config=self,
            embedding=embedding,
            transformer=transformer,
        )


class Decoder(LalamoModule[DecoderConfig]):
    embedding: EmbeddingBase
    transformer: Transformer

    @property
    def vocab_size(self) -> int:
        return self.embedding.vocab_size

    @property
    def activation_precision(self) -> DTypeLike:
        return self.embedding.activation_precision

    @eqx.filter_jit
    def __call__(
        self,
        token_ids: Int[Array, "batch suffix_tokens"],
        token_positions: Int[Array, "batch suffix_tokens"],
        state: State | None = None,
        return_updated_state: bool = False,
        return_activation_trace: bool = False,
        lengths_without_padding: Int[Array, " batch"] | None = None,
        forward_pass_mode: ForwardPassMode = ForwardPassMode.MULTI_TOKEN,
        forward_pass_config: DecoderForwardPassConfig | None = None,
    ) -> DecoderResult:
        if token_ids.ndim != 2:
            raise ValueError(
                f"token_ids must be a 2D array of size (batch_size, sequence_length), got {token_ids.shape}",
            )
        if token_positions.ndim != 2:
            raise ValueError(
                "token_positions must be a 2D array of size (batch_size, sequence_length),"
                f" got {token_positions.shape}",
            )

        inner_features = vmap(self.embedding.embed)(token_ids)

        transformer_result = self.transformer(
            inner_features=inner_features,
            token_positions=token_positions,
            state=state,
            return_updated_state=return_updated_state,
            return_layer_results=return_activation_trace,
            return_positional_embeddings=return_activation_trace,
            lengths_without_padding=lengths_without_padding,
            forward_pass_mode=forward_pass_mode,
            forward_pass_config=forward_pass_config,
        )

        logits = vmap_twice(self.embedding.readout)(transformer_result.outputs)

        if return_activation_trace:
            assert transformer_result.layer_results is not None

            activation_trace = DecoderActivationTrace(
                token_ids=token_ids,
                token_positions=token_positions,
                state=state,
                global_positional_embeddings=transformer_result.global_positional_embeddings,
                local_positional_embeddings=transformer_result.local_positional_embeddings,
                layer_results=transformer_result.layer_results,
                output_norm=transformer_result.outputs,
            )
        else:
            activation_trace = None

        return DecoderResult(
            logits=logits,
            updated_state=transformer_result.updated_state,
            activation_trace=activation_trace,
        )

    def init_static_state(self, batch_size: int, capacity: int) -> State:
        return self.transformer.init_static_state(batch_size, capacity)

    def export_weights(self) -> ParameterTree:
        return dict(
            embedding=self.embedding.export_weights(),
            transformer=self.transformer.export_weights(),
        )

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            embedding=self.embedding.import_weights(require_tree(weights["embedding"])),
            transformer=self.transformer.import_weights(require_tree(weights["transformer"])),
        )



import importlib.metadata
import json
from collections import ChainMap
from collections.abc import Callable
from contextlib import ExitStack
from dataclasses import dataclass, replace
from pathlib import Path
from typing import NamedTuple

import huggingface_hub
import jax.numpy as jnp
from jax import Array
from jaxtyping import DTypeLike
from tokenizers import Tokenizer

from lalamo.message_processor import MessageProcessor, MessageProcessorConfig
from lalamo.models import ClassifierModel, ClassifierModelConfig, GenerationConfig, LanguageModel, LanguageModelConfig
from lalamo.modules import Classifier, Decoder, LalamoModule
from lalamo.quantization import QuantizationMode
from lalamo.utils import process_chat_template

from .decoder_configs import ForeignClassifierConfig, ForeignConfig, ForeignLMConfig
from .huggingface_generation_config import HFGenerationConfig, _policy_from_hf_config
from .huggingface_tokenizer_config import HFTokenizerConfig
from .model_specs import REPO_TO_MODEL, FileSpec, ModelSpec, ModelType, UseCase
from .model_specs.common import JSONFieldSpec

__all__ = [
    "REPO_TO_MODEL",
    "DownloadingFileEvent",
    "FinishedDownloadingFileEvent",
    "InitializingModelEvent",
    "ModelMetadata",
    "ModelSpec",
    "ModelType",
    "StatusEvent",
    "download_file",
    "import_model",
]


LALAMO_VERSION = importlib.metadata.version("lalamo")


class DownloadingFileEvent(NamedTuple):
    file: FileSpec


class FinishedDownloadingFileEvent(NamedTuple):
    file: FileSpec


class InitializingModelEvent(NamedTuple):
    pass


class FinishedInitializingModelEvent(NamedTuple):
    pass


type StatusEvent = (
    DownloadingFileEvent | FinishedDownloadingFileEvent | InitializingModelEvent | FinishedInitializingModelEvent
)


@dataclass(frozen=True)
class ModelMetadata:
    toolchain_version: str
    vendor: str
    family: str
    name: str
    size: str
    quantization: QuantizationMode | None
    repo: str
    use_cases: tuple[UseCase, ...]
    model_type: ModelType
    model_config: LanguageModelConfig | ClassifierModelConfig
    grammar_start_tokens: tuple[str, ...]


def download_file(
    file_spec: FileSpec,
    model_repo: str,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> Path:
    if progress_callback is not None:
        progress_callback(DownloadingFileEvent(file_spec))
    result = huggingface_hub.hf_hub_download(
        repo_id=file_spec.repo or model_repo,
        local_dir=output_dir,
        filename=file_spec.filename,
    )
    if progress_callback is not None:
        progress_callback(FinishedDownloadingFileEvent(file_spec))
    return Path(result)


def list_weight_files(model_repo: str) -> list[FileSpec]:
    all_files = huggingface_hub.list_repo_files(model_repo)
    return [FileSpec(filename) for filename in all_files if filename.endswith(".safetensors")]


def download_weights(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> list[Path]:
    return [
        download_file(file_spec, model_spec.repo, output_dir, progress_callback)
        for file_spec in list_weight_files(model_spec.repo)
    ]


def download_config_file(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> Path:
    return download_file(model_spec.configs.model_config, model_spec.repo, output_dir, progress_callback)


class ImportResults(NamedTuple):
    model: LanguageModel | ClassifierModel
    metadata: ModelMetadata


def import_message_processor(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> MessageProcessor:
    tokenizer_file = download_file(model_spec.configs.tokenizer, model_spec.repo, output_dir, progress_callback)
    tokenizer_config_file = download_file(
        model_spec.configs.tokenizer_config,
        model_spec.repo,
        output_dir,
        progress_callback,
    )
    tokenizer_config = HFTokenizerConfig.from_json(tokenizer_config_file)

    if tokenizer_config.chat_template is None:
        match model_spec.configs.chat_template:
            case JSONFieldSpec(file_spec, field_name):
                json_file = download_file(file_spec, model_spec.repo, output_dir)
                with open(json_file) as file:
                    json_dict = json.load(file)
                prompt_template = json_dict[field_name]
            case FileSpec(_) as file_spec:
                chat_template_file = download_file(file_spec, model_spec.repo, output_dir)
                prompt_template = chat_template_file.read_text()
            case str() as template_string:
                prompt_template = template_string
            case None:
                raise ValueError("No chat template specified.")
    else:
        if model_spec.configs.chat_template is not None:
            raise ValueError("Conflicting chat template specifications.")
        prompt_template = tokenizer_config.chat_template
    prompt_template = process_chat_template(prompt_template)
    tokenizer = Tokenizer.from_file(str(tokenizer_file))

    added_tokens = tokenizer_config.added_tokens()
    added_special_tokens = [token for token in added_tokens if token.special]
    added_not_special_tokens = [token for token in added_tokens if not token.special]
    tokenizer.add_special_tokens(added_special_tokens)
    tokenizer.add_tokens(added_not_special_tokens)

    bos_token = tokenizer_config.bos_token
    eos_token = tokenizer_config.eos_token

    # If we were not able to identify bos/eos - they are probably somewhere else, so we check config.json
    if eos_token is None or bos_token is None:
        foreign_decoder_config_file = download_config_file(model_spec, output_dir, progress_callback)
        with open(foreign_decoder_config_file) as foreign_decoder_file:
            foreign_decoder_json = json.load(foreign_decoder_file)

        if bos_token is None:
            bos_token = foreign_decoder_json.get("bos_token_id")
        if eos_token is None:
            eos_token = foreign_decoder_json.get("eos_token_id")

    message_processor_config = MessageProcessorConfig(
        prompt_template=prompt_template,
        output_parser_regex=model_spec.output_parser_regex,
        system_role_name=model_spec.system_role_name,
        user_role_name=model_spec.user_role_name,
        assistant_role_name=model_spec.assistant_role_name,
        bos_token=bos_token,
        eos_token=eos_token,
    )
    return MessageProcessor(config=message_processor_config, tokenizer=tokenizer)


def _load_main_processing_module(
    model_spec: ModelSpec,
    precision: DTypeLike,
    foreign_config: ForeignConfig,
    progress_callback: Callable[[StatusEvent], None] | None = None,
    context_length: int | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
) -> LalamoModule:
    weights_paths = download_weights(model_spec, progress_callback=progress_callback)
    with ExitStack() as stack:
        weights_shards = []
        metadata_shards = []
        for weights_path in weights_paths:
            weights_shard, metadata_shard = stack.enter_context(model_spec.weights_type.load(weights_path, precision))
            weights_shards.append(weights_shard)
            metadata_shards.append(metadata_shard)
        weights_dict: ChainMap[str, Array] = ChainMap(*weights_shards)
        metadata_dict: ChainMap[str, str] = ChainMap(*metadata_shards)

        if progress_callback is not None:
            progress_callback(InitializingModelEvent())

        processing_module = foreign_config.load(
            context_length,
            precision,
            accumulation_precision,
            weights_dict,
            metadata_dict,
        )

    return processing_module


def _import_language_model(
    model_spec: ModelSpec,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> tuple[LanguageModel, LanguageModelConfig]:
    foreign_decoder_config_file = download_config_file(model_spec)
    foreign_decoder_config = model_spec.config_type.from_json(foreign_decoder_config_file)
    assert isinstance(foreign_decoder_config, ForeignLMConfig)

    if precision is None:
        precision = foreign_decoder_config.default_precision
    decoder = _load_main_processing_module(
        model_spec,
        precision,
        foreign_decoder_config,
        progress_callback,
        context_length,
        accumulation_precision,
    )
    assert isinstance(decoder, Decoder)

    if progress_callback is not None:
        progress_callback(FinishedInitializingModelEvent())

    message_processor = import_message_processor(model_spec)

    stop_token_ids = tuple(foreign_decoder_config.eos_token_ids)

    if isinstance(model_spec.configs.generation_config, GenerationConfig):
        generation_config = replace(model_spec.configs.generation_config, stop_token_ids=stop_token_ids)
    elif isinstance(model_spec.configs.generation_config, FileSpec):
        hf_generation_config_file = download_file(model_spec.configs.generation_config, model_spec.repo)
        hf_generation_config = HFGenerationConfig.from_json(hf_generation_config_file)
        generation_config = _policy_from_hf_config(hf_generation_config, stop_token_ids)
    else:
        generation_config = GenerationConfig(stop_token_ids)

    language_model_config = LanguageModelConfig(
        model_config=decoder.config,
        message_processor_config=message_processor.config,
        generation_config=generation_config,
    )

    language_model = LanguageModel(language_model_config, decoder, message_processor)
    return language_model, language_model_config


def _import_classifier(
    model_spec: ModelSpec,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> tuple[ClassifierModel, ClassifierModelConfig]:
    foreign_classifier_config_file = download_config_file(model_spec)
    foreign_classifier_config = model_spec.config_type.from_json(foreign_classifier_config_file)
    assert isinstance(foreign_classifier_config, ForeignClassifierConfig)

    if precision is None:
        precision = foreign_classifier_config.default_precision

    classifier = _load_main_processing_module(
        model_spec,
        precision,
        foreign_classifier_config,
        progress_callback,
        context_length,
        accumulation_precision,
    )
    assert isinstance(classifier, Classifier)

    if progress_callback is not None:
        progress_callback(FinishedInitializingModelEvent())

    message_processor = import_message_processor(model_spec)

    classifier_model_config = ClassifierModelConfig(
        model_config=classifier.config,
        message_processor_config=message_processor.config,
    )
    classifier_model = ClassifierModel(classifier_model_config, classifier, message_processor)
    return classifier_model, classifier_model_config


def import_model(
    model_spec: ModelSpec | str,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> ImportResults:
    if isinstance(model_spec, str):
        try:
            model_spec = REPO_TO_MODEL[model_spec]
        except KeyError as e:
            raise ValueError(f"Unknown model: {model_spec}") from e

    match model_spec.model_type:
        case ModelType.LANGUAGE_MODEL:
            model, config = _import_language_model(
                model_spec,
                context_length=context_length,
                precision=precision,
                accumulation_precision=accumulation_precision,
                progress_callback=progress_callback,
            )
        case ModelType.CLASSIFIER_MODEL:
            model, config = _import_classifier(
                model_spec,
                context_length=context_length,
                precision=precision,
                accumulation_precision=accumulation_precision,
                progress_callback=progress_callback,
            )

    metadata = ModelMetadata(
        toolchain_version=LALAMO_VERSION,
        vendor=model_spec.vendor,
        family=model_spec.family,
        name=model_spec.name,
        size=model_spec.size,
        quantization=model_spec.quantization,
        repo=model_spec.repo,
        use_cases=model_spec.use_cases,
        model_type=model_spec.model_type,
        model_config=config,
        grammar_start_tokens=model_spec.grammar_start_tokens,
    )
    return ImportResults(model, metadata)



import json
import struct
from collections.abc import Mapping
from dataclasses import dataclass
from io import BufferedReader, BufferedWriter
from typing import Any, ClassVar, Self

import cattrs
import jax
import jax.numpy as jnp
import numpy as np
from jaxtyping import Array

from lalamo.utils import LazyDict

SF2J = {
    "BOOL": jnp.dtype(jnp.bool_),
    "U8": jnp.dtype(jnp.uint8),
    "I8": jnp.dtype(jnp.int8),
    "I16": jnp.dtype(jnp.int16),
    "U16": jnp.dtype(jnp.uint16),
    "F16": jnp.dtype(jnp.float16),
    "BF16": jnp.dtype(jnp.bfloat16),
    "I32": jnp.dtype(jnp.int32),
    "U32": jnp.dtype(jnp.uint32),
    "F32": jnp.dtype(jnp.float32),
    "C64": jnp.dtype(jnp.complex64),
    "F64": jnp.dtype(jnp.float64),
    "I64": jnp.dtype(jnp.int64),
    "U64": jnp.dtype(jnp.uint64),
}

J2SF = {v: k for k, v in SF2J.items()}


@dataclass(frozen=True)
class SFTensorInfo:
    _converter: ClassVar[cattrs.Converter] = cattrs.Converter()
    _converter.register_structure_hook(jnp.dtype, lambda x, _: SF2J[x])
    _converter.register_unstructure_hook(jnp.dtype, lambda x: J2SF[x])

    dtype: jnp.dtype
    shape: tuple[int, ...]
    data_offsets: tuple[int, int]

    @property
    def start(self) -> int:
        return self.data_offsets[0]

    @property
    def end(self) -> int:
        return self.data_offsets[1]

    @property
    def size(self) -> int:
        return self.end - self.start

    @classmethod
    def from_dict(cls, obj: dict) -> Self:
        return cls._converter.structure(obj, cls)

    def to_dict(self) -> dict:
        return self._converter.unstructure(self)


def safe_read(fd: BufferedReader) -> tuple[dict[str, str] | None, LazyDict[str, Array]]:
    header_size = struct.unpack("<Q", fd.read(8))[0]
    header: dict[str, dict[str, Any]] = json.loads(fd.read(header_size))
    metadata: dict[str, str] | None = header.pop("__metadata__", None)
    data_offset = fd.tell()

    def _load_tensor(key: str) -> Array:
        info = SFTensorInfo.from_dict(header[key])
        fd.seek(data_offset + info.start)
        return jnp.asarray(np.fromfile(fd, info.dtype, info.size // info.dtype.itemsize)).reshape(info.shape)

    lazy_tensors = LazyDict(set(header.keys()), _load_tensor)
    return (metadata, lazy_tensors)


def safe_write(fd: BufferedWriter, tensors: Mapping[str, Array]) -> None:
    sorted_tensors = dict(sorted(tensors.items(), key=lambda x: (-x[1].dtype.alignment, x[0])))

    header_dict = {}
    offset = 0
    for key, tensor in sorted_tensors.items():
        assert offset % tensor.dtype.alignment == 0
        header_dict[key] = SFTensorInfo(tensor.dtype, tensor.shape, (offset, offset + tensor.nbytes)).to_dict()
        offset += tensor.nbytes

    data_alignment = max(8, next((t.dtype.alignment for t in sorted_tensors.values()), 1))
    header = json.dumps(header_dict).encode()
    header += b" " * (-len(header) % data_alignment)
    fd.write(struct.pack("<Q", len(header)) + header)

    for tensor in sorted_tensors.values():
        jax.device_get(tensor).tofile(fd)



from collections.abc import Callable

import jax
from jax import vmap
from jaxtyping import Array, Float

__all__ = [
    "apply_soft_capping",
    "vmap_twice",
]


def vmap_twice[F: Callable](
    func: F,
) -> F:
    return vmap(vmap(func, in_axes=0), in_axes=0)


def apply_soft_capping(
    values: Float[Array, "*"],
    soft_cap: float,
) -> Float[Array, "dst_tokens src_tokens"]:
    return jax.nn.tanh(values / soft_cap) * soft_cap



import os
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import cast

import jax
import jax.numpy as jnp
from jax._src.api import ShapeDtypeStruct
from jaxtyping import Array, DTypeLike

from lalamo.utils import MapDictValues, MapSequence

__all__ = [
    "DEFAULT_PRECISION",
    "ArrayLike",
    "LalamoWarning",
    "ParameterPath",
    "ParameterTree",
    "dummy_array",
    "flatten_parameters",
    "require_array",
    "require_tree",
    "unflatten_parameters",
]

DEFAULT_PRECISION: DTypeLike = jnp.bfloat16


class LalamoWarning(UserWarning):
    """Custom warning class for Lalamo-specific warnings."""


type ArrayLike = Array | ShapeDtypeStruct


type ParameterTree[ArrayType: ArrayLike] = (
    Mapping[str, ArrayType | ParameterTree[ArrayType]] | Sequence[ArrayType | ParameterTree[ArrayType]]
)


def require_array[ArrayType: ArrayLike](value: ArrayType | ParameterTree[ArrayType]) -> ArrayType:
    assert not isinstance(value, (Mapping, Sequence))
    return value


def require_tree[ArrayType: ArrayLike](value: ArrayType | ParameterTree[ArrayType]) -> ParameterTree[ArrayType]:
    assert not isinstance(value, (Array, ShapeDtypeStruct))
    return value


def dummy_array(shape: int | tuple[int, ...], dtype: DTypeLike) -> Array:
    if isinstance(shape, int):
        shape = (shape,)
    return cast("Array", ShapeDtypeStruct(shape=shape, dtype=dtype))


def flatten_parameters[ArrayType: ArrayLike](nested_parameters: ParameterTree[ArrayType]) -> dict[str, ArrayType]:
    result: dict[str, ArrayType] = {}
    if not isinstance(nested_parameters, Mapping):
        nested_parameters = {str(i): value for i, value in enumerate(nested_parameters)}
    for key, value in nested_parameters.items():
        value = cast("ArrayType | ParameterTree[ArrayType]", value)
        key_path = ParameterPath(key)
        if isinstance(value, (Array, ShapeDtypeStruct)):
            result[key_path] = cast("ArrayType", value)
        else:
            update: dict[str, ArrayType] = {
                str(key_path / subkey): subvalue for subkey, subvalue in flatten_parameters(value).items()
            }
            result.update(update)
    return result


type KeyTree = Mapping[str, str | KeyTree] | Sequence[str | KeyTree]


def _unflatten_keys(flat_keys: Mapping[str, str]) -> KeyTree:
    groups: dict[str, dict[str, str] | str] = defaultdict(dict)
    for subkey, full_key in flat_keys.items():
        match subkey.split(".", maxsplit=1):
            case [head]:
                groups[head] = full_key
            case [head, tail]:
                group = groups[head]
                assert isinstance(group, dict)
                group[tail] = full_key

    unflattened_groups: dict[str, KeyTree] = {}
    for subkey, group in groups.items():
        if isinstance(group, str):
            unflattened_groups[subkey] = group
        else:
            unflattened_groups[subkey] = _unflatten_keys(group)

    if any(key.isnumeric() for key in unflattened_groups):
        assert set(unflattened_groups.keys()) == set(map(str, range(len(unflattened_groups))))
        return [v for k, v in sorted(unflattened_groups.items(), key=lambda item: int(item[0]))]
    return unflattened_groups


def _recursive_map_dict[ArrayType: ArrayLike](
    key_tree: KeyTree | str,
    root_collection: Mapping[str, ArrayType],
) -> ParameterTree[ArrayType] | ArrayType:
    if isinstance(key_tree, str):
        return root_collection[key_tree]
    if isinstance(key_tree, Mapping):
        return MapDictValues(lambda subtree: _recursive_map_dict(subtree, root_collection), key_tree)
    if isinstance(key_tree, Sequence):
        return MapSequence(lambda subtree: _recursive_map_dict(subtree, root_collection), key_tree)


def unflatten_parameters[ArrayType: ArrayLike](flat_parameters: Mapping[str, ArrayType]) -> ParameterTree[ArrayType]:
    unflattened_keys = _unflatten_keys({k: k for k in flat_parameters})
    result = _recursive_map_dict(unflattened_keys, flat_parameters)
    assert not isinstance(result, (Array, ShapeDtypeStruct))
    return result


class ParameterPath(str):
    __slots__ = ()

    @property
    def components(self) -> tuple[str, ...]:
        return tuple(self.split("."))

    def __truediv__(self, other: str | int) -> "ParameterPath":
        if not self:
            return ParameterPath(str(other))
        return ParameterPath(self + "." + str(other))


def get_default_device_bytes() -> int | None:
    dynamic_allocate = False

    preallocate = os.getenv("XLA_PYTHON_CLIENT_PREALLOCATE", "")
    dynamic_allocate |= preallocate.strip().lower() in {"0", "false", "no", "off"}

    allocator = os.getenv("XLA_PYTHON_CLIENT_ALLOCATOR", "")
    dynamic_allocate |= allocator.strip().lower() in {"platform", "cuda_malloc_async"}

    if dynamic_allocate:
        return None

    memory_stats = jax.local_devices()[0].memory_stats()
    if memory_stats is None or "bytes_limit" not in memory_stats:
        return None

    # 500mb is seemingly the usually observed overhead
    memory_limit = memory_stats["bytes_limit"] - (500 * 1000 * 1000)

    return memory_limit


def get_usable_memory_from_bytes(limit_bytes: int) -> int:
    return int(limit_bytes * 0.95)



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Literal

import jax.numpy as jnp
from jaxtyping import DTypeLike

from lalamo.modules import (
    AttentionConfig,
    DecoderConfig,
    DenseMLPConfig,
    FullPrecisionLinearConfig,
    MLXQuantizedLinearConfig,
    MLXQuantizedTiedEmbeddingConfig,
    NormalizationConfig,
    SeparableCausalConvConfig,
    ShortConvConfig,
    SiLU,
    TiedEmbeddingConfig,
    TransformerConfig,
    TransformerLayerConfig,
    UnscaledRoPEConfig,
    UntiedEmbeddingConfig,
    UpcastMode,
)
from lalamo.quantization import QuantizationMode

from .common import HuggingFaceLMConfig


@dataclass(frozen=True)
class QuantizationConfig:
    group_size: int
    bits: int


@dataclass(frozen=True)
class HFLFM2Config(HuggingFaceLMConfig):
    architectures: list[Literal["Lfm2ForCausalLM"]]
    block_auto_adjust_ff_dim: bool
    block_dim: int
    block_ff_dim: int
    block_ffn_dim_multiplier: float
    block_mlp_init_scale: float
    block_multiple_of: int
    block_norm_eps: float
    block_out_init_scale: float
    block_use_swiglu: bool
    block_use_xavier_init: bool
    bos_token_id: int
    conv_L_cache: int  # noqa: N815
    conv_bias: bool
    conv_dim: int
    conv_use_xavier_init: bool
    eos_token_id: int
    hidden_size: int
    initializer_range: float
    max_position_embeddings: int
    model_type: Literal["lfm2"]
    norm_eps: float
    num_attention_heads: int
    num_heads: int
    num_hidden_layers: int
    num_key_value_heads: int
    pad_token_id: int
    rope_theta: float
    transformers_version: str
    use_cache: bool
    use_pos_enc: bool
    vocab_size: int

    dtype: Literal["bfloat16", "float16", "float32"] | None = None
    torch_dtype: Literal["bfloat16", "float16", "float32"] | None = None
    intermediate_size: int | None = None
    conv_dim_out: int | None = None
    layer_types: list[Literal["conv", "full_attention"]] | None = None
    full_attn_idxs: list[int] | None = None
    tie_embedding: bool = True
    theta: float | None = None

    quantization: QuantizationConfig | None = None
    quantization_config: QuantizationConfig | None = None

    @property
    def default_precision(self) -> DTypeLike:
        assert self.dtype is not None or self.torch_dtype is not None, (
            "at least one of dtype or torch_dtype must be specified"
        )

        return jnp.dtype(self.dtype or self.torch_dtype)

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        metadata_dict: Mapping[str, str],  # noqa: ARG002
    ) -> DecoderConfig:
        assert self.num_attention_heads == self.num_heads

        if self.quantization_config is not None:
            assert self.tie_embedding

            embedding_config = MLXQuantizedTiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                group_size=self.quantization_config.group_size,
                embedding_quantization_mode=QuantizationMode.from_num_bits(self.quantization_config.bits),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )
        elif self.tie_embedding:
            embedding_config = TiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                precision=activation_precision,
            )
        else:
            embedding_config = UntiedEmbeddingConfig(
                input_scale=None,
                logit_soft_cap=None,
                precision=activation_precision,
            )

        rope_config = UnscaledRoPEConfig(
            precision=activation_precision,
            base=self.rope_theta,
            max_sequence_length=context_length or self.max_position_embeddings,
        )

        if self.quantization_config is None:
            linear_config = FullPrecisionLinearConfig(activation_precision)
        else:
            linear_config = MLXQuantizedLinearConfig(
                group_size=self.quantization_config.group_size,
                weight_quantization_mode=QuantizationMode.from_num_bits(self.quantization_config.bits),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )

        block_norm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.block_norm_eps,
            scale_offset=None,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
            subtract_mean=False,
        )

        attention_config = AttentionConfig(
            qkv_projection_config=linear_config,
            out_projection_config=linear_config,
            query_norm_config=block_norm_config,
            key_norm_config=block_norm_config,
            num_heads=self.num_attention_heads,
            num_groups=self.num_key_value_heads,
            head_dim=self.hidden_size // self.num_heads,
            is_causal=True,
            scale=None,
            sliding_window_size=None,
            logit_soft_cap=None,
            has_sinks=False,
            has_qkv_biases=False,
            has_out_biases=False,
        )

        short_conv_config = ShortConvConfig(
            in_projection_config=linear_config,
            conv_config=SeparableCausalConvConfig(activation_precision, has_biases=self.conv_bias),
            out_projection_config=linear_config,
            kernel_size=self.conv_L_cache,
        )

        mlp_config = DenseMLPConfig(
            linear_config=linear_config,
            activation=SiLU(),
            has_up_biases=False,
            has_down_biases=False,
            up_clipping=None,
            gate_clipping=None,
        )

        if self.layer_types is not None:
            layer_types = self.layer_types
        elif self.full_attn_idxs is not None:
            layer_types = [
                "full_attention" if i in self.full_attn_idxs else "conv" for i in range(self.num_hidden_layers)
            ]
        else:
            raise RuntimeError("Either layer_types or full_attn_idxs must be present.")

        layer_configs = [
            TransformerLayerConfig(
                pre_mixer_norm_config=block_norm_config,
                mixer_config={"conv": short_conv_config, "full_attention": attention_config}[layer_type],
                post_mixer_norm_config=None,
                pre_mlp_norm_config=block_norm_config,
                mlp_config=mlp_config,
                post_mlp_norm_config=None,
            )
            for layer_type in layer_types
        ]

        output_norm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.norm_eps,
            scale_offset=None,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
            subtract_mean=False,
        )

        if not self.block_auto_adjust_ff_dim:
            hidden_dim = self.intermediate_size or self.block_ff_dim
        else:
            hidden_dim_adjusted = self.block_ff_dim * self.block_ffn_dim_multiplier * (2 / 3)
            hidden_dim = int(
                (hidden_dim_adjusted + self.block_multiple_of - 1) // self.block_multiple_of * self.block_multiple_of,
            )

        transformer_config = TransformerConfig(
            global_rope_config=rope_config,
            local_rope_config=None,
            layer_configs=tuple(layer_configs),
            output_norm_config=output_norm_config,
            model_dim=self.hidden_size,
            hidden_dim=hidden_dim,
            context_length=context_length or self.max_position_embeddings,
        )

        return DecoderConfig(
            embedding_config=embedding_config,
            transformer_config=transformer_config,
            vocab_size=self.vocab_size,
        )



import random
import re
import shutil
import sys
from contextlib import ExitStack
from dataclasses import dataclass, field
from functools import partial
from itertools import islice
from pathlib import Path
from typing import Annotated

import jax.profiler
import requests
import thefuzz.process
from click import Context as ClickContext
from click import Parameter as ClickParameter
from click import ParamType
from rich import box
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TaskID,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.prompt import Confirm
from rich.table import Table
from typer import Argument, Context, Exit, Option, Typer

from lalamo.commands import (
    CollectTracesCallbacks,
    ConversionCallbacks,
    EstimateBatchsizeCallbacks,
    GenerateRepliesCallbacks,
    Precision,
    PullCallbacks,
    TraceCallbacks,
    TrainCallbacks,
    _suggest_similar_models,
)
from lalamo.commands import collect_traces as _collect_traces
from lalamo.commands import convert as _convert
from lalamo.commands import estimate_batchsize as _estimate_batchsize
from lalamo.commands import generate_replies as _generate_replies
from lalamo.commands import pull as _pull
from lalamo.commands import trace as _trace
from lalamo.commands import train as _train
from lalamo.common import (
    get_default_device_bytes,
    get_usable_memory_from_bytes,
)
from lalamo.data.lalamo_completions import LalamoCompletion
from lalamo.message_processor import UserMessage
from lalamo.model_import import REPO_TO_MODEL, ModelSpec
from lalamo.model_import.common import FileSpec
from lalamo.model_import.remote_registry import RegistryModel, RegistryModelFile, fetch_available_models
from lalamo.models import ClassifierModelConfig, LanguageModelConfig
from lalamo.models.common import BatchSizesComputedEvent
from lalamo.speculator.ngram import NGramSpeculator
from lalamo.speculator.utils import test_speculator

SCRIPT_NAME = Path(sys.argv[0]).name

DEFAULT_OUTPUT_DIR = Path("models")


console = Console()
err_console = Console(stderr=True)
app = Typer(
    rich_markup_mode="rich",
    add_completion=False,
    pretty_exceptions_show_locals=False,
)


class ModelParser(ParamType):
    name: str = "Huggingface Model Repo"

    def convert(self, value: str, param: ClickParameter | None, ctx: ClickContext | None) -> ModelSpec:
        result = REPO_TO_MODEL.get(value)
        if result is None:
            closest_repo = _closest_repo(value, list(REPO_TO_MODEL))
            error_message_parts = [
                f'"{value}".',
            ]
            if closest_repo:
                error_message_parts.append(
                    f' Perhaps you meant "{closest_repo}"?',
                )
            error_message_parts.append(
                f"\n\nUse the `{SCRIPT_NAME} list-models` command to see the list of currently supported models.",
            )
            error_message = "".join(error_message_parts)
            return self.fail(error_message, param, ctx)
        return result


class RemoteModelParser(ParamType):
    name: str = "Pre-converted Model"

    def convert(self, value: str, param: ClickParameter | None, ctx: ClickContext | None) -> "RegistryModel":
        try:
            available_models = fetch_available_models()
        except (requests.RequestException, ValueError) as e:
            error_message = f"Failed to fetch model list from SDK. Check your internet connection.\n\nError: {e}"
            return self.fail(error_message, param, ctx)

        repo_to_model = {m.repo_id: m for m in available_models}
        model_spec = repo_to_model.get(value)
        if model_spec is None:
            closest_repo = _closest_repo(value, list(repo_to_model))
            if closest_repo:
                model_spec = repo_to_model[closest_repo]

        if model_spec is None:
            suggestions = _suggest_similar_models(value, available_models)
            error_message = f'Model "{value}" not found.'
            if suggestions:
                error_message += "\n\nDid you mean one of these?\n" + "\n".join(f"  - {s}" for s in suggestions)
            return self.fail(error_message, param, ctx)

        return model_spec


def _closest_repo(query: str, repo_ids: list[str], min_score: float = 80) -> str | None:
    if not repo_ids:
        return None
    (closest_match, score), *_ = thefuzz.process.extract(query, repo_ids)
    if closest_match and score >= min_score:
        return closest_match
    return None


def _error(message: str) -> None:
    panel = Panel(message, box=box.ROUNDED, title="Error", title_align="left", border_style="red")
    err_console.print(panel)
    raise Exit(1)


@app.command(help="Chat with a converted model.")
def chat(
    model_path: Annotated[
        Path,
        Argument(
            help="Path to the model directory.",
            metavar="MODEL_PATH",
        ),
    ],
    message: Annotated[
        str | None,
        Option(
            help="Message for non-interactive mode",
            show_default="None, run interactively",
        ),
    ] = None,
) -> None:
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=err_console,
        transient=True,
    ) as progress:
        loading_task = progress.add_task("🚀 [cyan]Loading model...[/cyan]")
        model = LanguageModelConfig.load_model(model_path)
        progress.remove_task(loading_task)
        warmup_task = progress.add_task("🔥 Warming up compilation cache...")
        list(model.stream_reply_text([UserMessage("")], max_output_length=1))
        progress.remove_task(warmup_task)

    if message is None:
        console.print(f"🤖 Chatting with [blue]{model_path}[/blue]:")

        messages = []
        while True:
            user_text = console.input("[cyan]user> [/cyan]")
            user_message = UserMessage(user_text)
            messages.append(user_message)

            console.print("[red]assistant> [/red]", end="")
            model_response_tokens = []
            for token in model.stream_reply_text(messages):
                console.print(token, end="")
                model_response_tokens.append(token)
            console.print()
            model_response_text = "".join(model_response_tokens)
            messages.append(model.message_processor.parse_response(model_response_text))
    else:
        for token in model.stream_reply_text([UserMessage(message)]):
            console.print(token, end="")
        console.print()


@app.command(help="Classify given message with a Classifier type of model.")
def classify(
    model_path: Annotated[
        Path,
        Argument(
            help="Path to the model directory.",
            metavar="MODEL_PATH",
        ),
    ],
) -> None:
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        loading_task = progress.add_task("🚀 [cyan]Loading model...[/cyan]")
        model = ClassifierModelConfig.load_model(model_path)
        progress.remove_task(loading_task)
        warmup_task = progress.add_task("🔥 Warming up...")
        model.classify_chat([UserMessage(content="warmup message")])
        progress.remove_task(warmup_task)
    console.print(f"🤖 Classifying input with [blue]{model_path}[/blue]:")
    while True:
        user_text = console.input("[cyan]user> [/cyan]")
        user_message = UserMessage(user_text)

        console.print("[red]assistant> [/red]", end="")
        result = model.classify_chat([user_message])
        for label, confidence in result.items():
            console.print(f"{label} : {confidence}", end="")
        console.print()


@dataclass
class CliConversionCallbacks(ConversionCallbacks):
    overwrite: bool = False

    stack: ExitStack = field(default_factory=ExitStack)
    progress: Progress | None = None
    downloading_tasks: dict[FileSpec, TaskID] = field(default_factory=dict)
    initializing_task: TaskID | None = None
    saving_task: TaskID | None = None

    def started(self) -> None:
        conversion_strs = [
            f"🚀 Converting [cyan]{self.model_spec.name}[/cyan] by [cyan]{self.model_spec.vendor}[/cyan]",
        ]
        if self.precision is not None:
            conversion_strs.append(
                f" and converting floating-point weights into [cyan]{self.precision.name.lower()}[/cyan] precision",
            )
        conversion_strs.append(".")
        console.print("".join(conversion_strs))

        self.progress = self.stack.enter_context(
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
            ),
        )

    def output_dir_exists(self) -> None:
        if not self.overwrite and not Confirm().ask(
            rf"⚠️ Output directory [cyan]{self.output_dir}[/cyan] already exists."
            r" Do you want to overwrite it?",
        ):
            raise Exit

        shutil.rmtree(self.output_dir)

    def downloading(self, file_spec: FileSpec) -> None:
        assert self.progress is not None

        self.downloading_tasks[file_spec] = self.progress.add_task(f"Retrieving {file_spec.filename}...")

    def finished_downloading(self, file_spec: FileSpec) -> None:
        assert self.progress is not None

        self.progress.remove_task(self.downloading_tasks[file_spec])

    def initializing_model(self) -> None:
        assert self.progress is not None

        self.initializing_task = self.progress.add_task("Initializing model...")

    def finished_initializing_model(self) -> None:
        assert self.progress is not None
        assert self.initializing_task is not None

        self.progress.remove_task(self.initializing_task)

    def saving_model(self) -> None:
        assert self.progress is not None

        self.saving_task = self.progress.add_task(f"💾 Saving the model to {self.output_dir}")

    def finished_saving_model(self) -> None:
        assert self.progress is not None
        assert self.saving_task is not None

        self.progress.remove_task(self.saving_task)
        self.stack.close()
        console.print(f"🧑‍🍳 Model successfully cooked and saved to [cyan]`{self.output_dir}`[/cyan]!")


@dataclass
class CliPullCallbacks(PullCallbacks):
    stack: ExitStack = field(default_factory=ExitStack)
    progress: Progress | None = None
    downloading_tasks: dict[RegistryModelFile, TaskID] = field(default_factory=dict)

    def started(self) -> None:
        console.print(f"📦 Pulling [cyan]{self.model_spec.name}[/cyan] by [cyan]{self.model_spec.vendor}[/cyan]")

        self.progress = self.stack.enter_context(
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
            ),
        )

    def output_dir_exists(self) -> None:
        if not self.overwrite and not Confirm().ask(
            rf"⚠️ Output directory [cyan]{self.output_dir}[/cyan] already exists."
            r" Do you want to overwrite it?",
        ):
            raise Exit

        shutil.rmtree(self.output_dir)

    def downloading(self, file_spec: RegistryModelFile) -> None:
        assert self.progress is not None

        self.downloading_tasks[file_spec] = self.progress.add_task(f"⬇️  Downloading {file_spec.name}...")

    def finished_downloading(self, file_spec: RegistryModelFile) -> None:
        assert self.progress is not None

        self.progress.remove_task(self.downloading_tasks[file_spec])

    def finished(self) -> None:
        assert self.progress is not None

        self.stack.close()
        console.print(f"🎉 Model successfully pulled to [cyan]{self.output_dir}[/cyan]!")


@app.command(help="Convert the model for use with the Uzu inference engine.")
def convert(
    model_repo: Annotated[
        ModelSpec,
        Argument(
            help=(
                "Huggingface model repo. Example: [cyan]'meta-llama/Llama-3.2-1B-Instruct'[/cyan]."
                "\n\n\n\n"
                f"You can use the [cyan]`{SCRIPT_NAME} list-models`[/cyan] command to get a list of supported models."
            ),
            click_type=ModelParser(),
            show_default=False,
            metavar="MODEL_REPO",
            autocompletion=lambda: list(REPO_TO_MODEL),
        ),
    ],
    precision: Annotated[
        Precision | None,
        Option(
            help="Precision to use for activations and non-quantized weights.",
            show_default="Native precision of the model",
        ),
    ] = None,
    output_dir: Annotated[
        Path | None,
        Option(
            help="Directory to save the converted model to.",
            show_default="Saves the converted model in the `models/<model_name>` directory",
        ),
    ] = None,
    context_length: Annotated[
        int | None,
        Option(
            help="Maximum supported context length. Used to precompute positional embeddings.",
            show_default="Model's native maximum context length.",
        ),
    ] = None,
    overwrite: Annotated[
        bool,
        Option(
            help="Overwrite existing model files.",
        ),
    ] = False,
) -> None:
    if output_dir is None:
        output_dir = DEFAULT_OUTPUT_DIR / model_repo.name

    _convert(
        model_repo,
        output_dir,
        precision,
        context_length,
        partial(CliConversionCallbacks, overwrite=overwrite),
    )


@app.command(help="Pull a pre-converted model from the SDK repository.")
def pull(
    model_spec: Annotated[
        RegistryModel,
        Argument(
            help=(
                "Model repository ID from the pre-converted catalog. "
                "Example: [cyan]'meta-llama/Llama-3.2-1B-Instruct'[/cyan]. "
                "Fuzzy matching is supported for typos and partial names."
            ),
            click_type=RemoteModelParser(),
            show_default=False,
            metavar="MODEL_IDENTIFIER",
        ),
    ],
    output_dir: Annotated[
        Path | None,
        Option(
            help="Directory to save the pulled model to.",
            show_default="Saves the pulled model in the `models/<model_name>` directory",
        ),
    ] = None,
    overwrite: Annotated[
        bool,
        Option(
            help="Overwrite existing model files without prompting.",
        ),
    ] = False,
) -> None:
    if output_dir is None:
        output_dir = DEFAULT_OUTPUT_DIR / model_spec.name

    _pull(
        model_spec,
        output_dir,
        partial(CliPullCallbacks),
        overwrite=overwrite,
    )


@dataclass
class CliTraceCallbacks(TraceCallbacks):
    overwrite: bool = False

    stack: ExitStack = field(default_factory=ExitStack)
    progress: Progress | None = None
    loading_task: TaskID | None = None
    tracing_task: TaskID | None = None
    saving_task: TaskID | None = None

    def output_exists(self) -> None:
        if not self.overwrite and not Confirm().ask(
            rf"⚠️ Output [cyan]{self.output_path}[/cyan] already exists."
            r" Do you want to overwrite it?",
        ):
            raise Exit

        self.output_path.unlink()

    def started(self) -> None:
        console.print(f"🔍 Tracing [cyan]{self.model_path}[/cyan]")

        self.progress = self.stack.enter_context(
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
            ),
        )

    def loading_model(self) -> None:
        assert self.progress is not None

        self.loading_task = self.progress.add_task("🧠 Loading model...")

    def finished_loading_model(self) -> None:
        assert self.progress is not None
        assert self.loading_task is not None

        self.progress.remove_task(self.loading_task)

    def tracing_model(self) -> None:
        assert self.progress is not None

        self.tracing_task = self.progress.add_task("🔍 Recording trace...")

    def finished_tracing_model(self) -> None:
        assert self.progress is not None
        assert self.tracing_task is not None

        self.progress.remove_task(self.tracing_task)

    def saving_trace(self) -> None:
        assert self.progress is not None

        self.saving_task = self.progress.add_task(f"💾 Saving trace to {self.output_path}")

    def finished_saving_trace(self) -> None:
        assert self.progress is not None
        assert self.saving_task is not None

        self.progress.remove_task(self.saving_task)
        self.stack.close()
        console.print(f"💾 Trace saved to [cyan]{self.output_path}[/cyan]")


@app.command(help="Trace a model.")
def trace(
    model_path: Annotated[
        Path,
        Argument(
            help="Path to the model directory.",
            metavar="MODEL_PATH",
        ),
    ],
    output_path: Annotated[
        Path | None,
        Option(
            help="Path to save the trace to.",
            show_default="${MODEL_PATH}/traces.safetensors",
        ),
    ] = None,
    overwrite: Annotated[
        bool,
        Option(
            help="Overwrite existing trace file.",
        ),
    ] = False,
    message: Annotated[
        str | None,
        Option(
            help="Text message to use as prompt when recording trace",
        ),
    ] = None,
) -> None:
    if output_path is None:
        output_path = model_path / "traces.safetensors"

    messages = None if message is None else [UserMessage(content=message)]

    _trace(
        model_path,
        output_path,
        messages,
        partial(CliTraceCallbacks, overwrite=overwrite),
    )


def _model_size_string_to_int(
    size_str: str,
    _regex: re.Pattern = re.compile(r"(?P<number>(\d+)(\.\d*)?)(?P<suffix>[KMBT])"),
) -> float:
    match = _regex.match(size_str)
    factors = {
        "K": 1000**1,
        "M": 1000**2,
        "B": 1000**3,
        "T": 1000**4,
    }
    if match:
        return float(match.group("number")) * factors[match.group("suffix")]
    raise ValueError(f"Invalid size string: {size_str}")


@app.command(help="List the supported models.")
def list_models(
    plain: Annotated[
        bool,
        Option(
            help="Only list repo names without fancy formatting.",
        ),
    ] = False,
) -> None:
    sorted_specs = sorted(
        REPO_TO_MODEL.values(),
        key=lambda spec: (
            spec.vendor.lower(),
            spec.family.lower(),
            _model_size_string_to_int(spec.size),
            spec.name.lower(),
        ),
    )

    if plain:
        for spec in sorted_specs:
            console.print(spec.repo)
        return

    table = Table(
        show_header=True,
        header_style="bold",
        show_lines=True,
        box=box.ROUNDED,
    )
    table.add_column("Vendor", justify="left", style="magenta")
    table.add_column("Family", justify="left", style="magenta", no_wrap=True)
    table.add_column("Size", justify="right", style="magenta")
    table.add_column("Quant", justify="left", style="magenta")
    table.add_column("Repo", justify="left", style="cyan", no_wrap=True)
    for spec in sorted_specs:
        table.add_row(
            spec.vendor,
            spec.family,
            spec.size,
            str(spec.quantization),
            spec.repo,
        )
    console.print(table)


@dataclass
class CliGenerateRepliesCallbacks(GenerateRepliesCallbacks):
    stack: ExitStack = field(default_factory=ExitStack)
    progress: Progress | None = None
    loading_task: TaskID | None = None
    estimating_task: TaskID | None = None
    generation_task: TaskID | None = None

    def loading_model(self) -> None:
        self.progress = self.stack.enter_context(
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                MofNCompleteColumn(),
                TimeElapsedColumn(),
                transient=True,
            ),
        )
        self.loading_task = self.progress.add_task("🧠 [cyan]Loading model...[/cyan]", total=None)

    def finished_loading_model(self) -> None:
        assert self.progress is not None
        assert self.loading_task is not None
        self.progress.remove_task(self.loading_task)

    def loading_dataset(self) -> None:
        assert self.progress is not None
        self.loading_task = self.progress.add_task("🗂️ [cyan]Loading dataset...[/cyan]", total=None)

    def finished_loading_dataset(self) -> None:
        assert self.progress is not None
        assert self.loading_task is not None
        self.progress.remove_task(self.loading_task)

    def estimating_batchsize(self, sequence_length: int, lo: int, hi: int | None) -> None:
        assert self.progress is not None
        hi_str = str(hi) if hi is not None else "?"
        description = (
            f"📐 [cyan]Computing batch size for the prompt length of {sequence_length}... ({lo}..{hi_str})[/cyan]"
        )
        if self.estimating_task is None:
            self.estimating_task = self.progress.add_task(description)
        else:
            self.progress.update(self.estimating_task, description=description)

    def batch_sizes_estimated(self) -> None:
        assert self.progress is not None
        if self.estimating_task is None:
            self.estimating_task = self.progress.add_task(
                "📐 [cyan]Estimating the best batch sizes...[/cyan]",
                total=None,
            )

    def batch_sizes_computed(self, event: BatchSizesComputedEvent) -> None:
        assert self.progress is not None
        if self.estimating_task is not None:
            self.progress.remove_task(self.estimating_task)
            self.estimating_task = None
        output_console = self.progress.console if self.progress is not None else console
        for info in event.batch_sizes:
            output_console.print(
                f"Prefix length {info.prefix_length} has {info.num_elements} elements, "
                f"with batchsize of {info.batch_size}",
            )
        self.generation_task = self.progress.add_task(
            "🔮 [cyan]Generating replies...[/cyan]",
            total=self.total_rows,
        )

    def generation_progress(self, rows_processed: int) -> None:
        assert self.progress is not None
        assert self.generation_task is not None
        self.progress.update(self.generation_task, completed=rows_processed + 1)

    def finished_generation(self) -> None:
        assert self.progress is not None
        assert self.generation_task is not None
        self.progress.update(self.generation_task, description="✅ Completed")
        self.stack.close()
        console.print(f"💾 Replies saved to [cyan]{self.output_path}[/cyan]")


@app.command(help="Generate replies for conversations in a parquet file.")
def generate_replies(
    model_path: Annotated[
        Path,
        Argument(
            help="Path to the model directory.",
            metavar="MODEL_PATH",
        ),
    ],
    dataset_path: Annotated[
        Path,
        Argument(
            help="Path to the input parquet file with conversations.",
            metavar="DATASET_PATH",
        ),
    ],
    output_path: Annotated[
        Path,
        Option(
            help="Path to save the output parquet file.",
        ),
    ],
    vram_gb: Annotated[
        int | None,
        Option(
            help="Maximum VRAM in GB. Batch sizes are estimated automatically.",
            show_default="max on default device",
        ),
    ] = None,
    max_output_length: Annotated[
        int,
        Option(help="Maximum number of tokens to generate per reply."),
    ] = 8192,
    batch_size: Annotated[
        int | None,
        Option(help="Fixed batch size to use, skipping automatic estimation."),
    ] = None,
) -> None:
    if batch_size is not None and vram_gb is not None:
        err_console.print("Cannot use both --batch-size and --vram-gb")
        raise Exit(1)

    max_vram: int | None = None
    if batch_size is None:
        if vram_gb is not None:
            mem_bytes = vram_gb * 1000 * 1000 * 1000
        elif (mem_bytes := get_default_device_bytes()) is None:
            err_console.print("Cannot get the default device's memory stats, use --vram-gb or --batch-size")
            raise Exit(1)

        max_vram = mem_bytes

    _generate_replies(
        model_path,
        dataset_path,
        output_path,
        max_vram,
        max_output_length,
        batch_size,
        CliGenerateRepliesCallbacks,
    )


speculator_app = Typer()
app.add_typer(speculator_app, name="speculator", help="Train a speculator for a model.")


@dataclass
class CliEstimateBatchsizeCallbacks(EstimateBatchsizeCallbacks):
    stack: ExitStack = field(default_factory=ExitStack)
    loading_task: TaskID | None = None
    estimating_task: TaskID | None = None

    def loading_model(self) -> None:
        self.progress = self.stack.enter_context(
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
            ),
        )
        self.loading_task = self.progress.add_task("[cyan]Loading model...[/cyan]")

    def finished_loading_model(self) -> None:
        assert self.loading_task is not None
        self.progress.remove_task(self.loading_task)

    def estimating_batchsize(self, lo: int, hi: int | None) -> None:
        hi_str = str(hi) if hi is not None else "?"
        description = f"[cyan]Estimating batch size... ({lo}..{hi_str})[/cyan]"
        if self.estimating_task is None:
            self.estimating_task = self.progress.add_task(description)
        else:
            self.progress.update(self.estimating_task, description=description)

    def finished_estimating_batchsize(self, batchsize: int) -> None:
        if self.estimating_task is not None:
            self.progress.remove_task(self.estimating_task)
        self.stack.close()
        console.print(f"Found maximum batch size: [cyan]{batchsize}[/cyan]")


@speculator_app.command(help="Estimate maximum batch size at which a model can be run.")
def estimate_batchsize(
    model_path: Annotated[
        Path,
        Argument(
            help="Path to the model directory",
            metavar="MODEL_PATH",
        ),
    ],
    max_input_length: Annotated[
        int,
        Option(help="Max input length of a model."),
    ] = 1024,
    max_output_length: Annotated[
        int,
        Option(help="Max output length of a model."),
    ] = 1024,
    num_logits_per_token: Annotated[
        int,
        Option(help="Number of top logits that will be recorded."),
    ] = 8,
    vram_gb: Annotated[
        int | None,
        Option(
            help="Maximum vram size in gb allowed.",
            show_default="max on default device",
        ),
    ] = None,
) -> None:
    if vram_gb is not None:
        # note that in practice GPUs use GiB in their docs, e.g. H100 actually has 85GB of memory
        mem_bytes = vram_gb * 1000 * 1000 * 1000
    elif (mem_bytes := get_default_device_bytes()) is None:
        err_console.print("Cannot get the default device's memory stats, use --vram-gb")
        raise Exit(1)

    usable_mem = get_usable_memory_from_bytes(mem_bytes)

    callbacks_type = CliEstimateBatchsizeCallbacks

    _estimate_batchsize(
        model_path,
        usable_mem,
        max_input_length,
        max_output_length,
        num_logits_per_token,
        callbacks_type,
    )


@dataclass
class CliCollectTracesCallbacks(CollectTracesCallbacks):
    stack: ExitStack = field(default_factory=ExitStack)
    live: Live | None = None
    loading_task: TaskID | None = None
    inference_task: TaskID | None = None

    def loading_model(self) -> None:
        self.live = self.stack.enter_context(Live(refresh_per_second=10))
        self.progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        )
        self.live.update(self.progress, refresh=True)
        self.loading_task = self.progress.add_task("🧠 [cyan]Loading model...[/cyan]")

    def finished_loading_model(self) -> None:
        assert self.loading_task is not None
        self.progress.remove_task(self.loading_task)

    def loading_dataset(self) -> None:
        self.loading_task = self.progress.add_task("🗂️ [cyan]Loading dataset...[/cyan]")

    def finished_loading_dataset(self) -> None:
        assert self.loading_task is not None
        assert self.live is not None
        self.progress.remove_task(self.loading_task)
        self.progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
        )
        self.live.update(self.progress, refresh=True)
        self.inference_task = self.progress.add_task(
            "🔮 [cyan]Running inference...[/cyan]",
            total=self.num_tokens_to_generate,
        )

    def inference_progress(self, tokens_generated: int) -> None:
        assert self.inference_task is not None
        self.progress.update(self.inference_task, completed=tokens_generated)

    def finished_inference(self) -> None:
        assert self.inference_task is not None
        self.progress.update(self.inference_task, description="✅ Completed")
        self.stack.close()


@speculator_app.command(help="Run model inference and collect traces for speculator training")
def collect_traces(
    model_path: Annotated[
        Path,
        Argument(
            help="Path to the model directory",
            metavar="MODEL_PATH",
        ),
    ],
    dataset_path: Annotated[
        Path,
        Argument(
            help="Path to the dataset with prompts",
            metavar="DATASET_PATH",
        ),
    ],
    output_path: Annotated[
        Path,
        Option(
            help="File to save the trace to",
            metavar="OUTPUT_PATH",
        ),
    ],
    num_logits_per_token: Annotated[
        int,
        Option(help="Record logits for this number of most probable tokens"),
    ] = 8,
    max_input_length: Annotated[
        int,
        Option(help="Filter prompts that have more than this number of tokens in context"),
    ] = 1024,
    max_output_length: Annotated[
        int,
        Option(help="Maximum number of tokens to generate in one completion"),
    ] = 1024,
    batch_size: Annotated[
        int,
        Option(help="Number of sequences in one batch"),
    ] = 1,
    num_tokens_to_generate: Annotated[
        int | None,
        Option(
            help="Exit early after generating this number of output tokens",
            show_default="all",
        ),
    ] = None,
) -> None:
    _collect_traces(
        model_path,
        dataset_path,
        output_path,
        num_logits_per_token,
        max_input_length,
        max_output_length,
        batch_size,
        num_tokens_to_generate,
        CliCollectTracesCallbacks,
    )


@speculator_app.command(help="View model inference traces")
def view_traces(
    trace_path: Annotated[
        Path,
        Argument(
            help="File of inference traces to view.",
            metavar="TRACE_PATH",
        ),
    ],
    model_path: Annotated[
        Path,
        Argument(
            help="Path to the model directory for detokenization.",
            metavar="MODEL_PATH",
        ),
    ],
    num_completions: Annotated[
        int | None,
        Option(
            help="Number of completions to show.",
        ),
    ] = None,
) -> None:
    model = LanguageModelConfig.load_model(model_path)

    with open(trace_path, "rb") as trace_fd:
        traces = LalamoCompletion.deserialize_many(trace_fd)

        table = Table(
            show_lines=True,
            box=box.ROUNDED,
        )
        table.add_column("Prefix")
        table.add_column("Completion")

        from rich.text import Text

        for completion in islice(traces, num_completions):
            detokenized_prefix = model.message_processor.detokenize(completion.prefix_token_ids)
            detokenized_completion = model.message_processor.detokenize(completion.completion_token_ids)
            table.add_row(Text(detokenized_prefix), Text(detokenized_completion))

        console.print(table)


@dataclass
class CliTrainCallbacks(TrainCallbacks):
    stack: ExitStack = field(default_factory=ExitStack)
    training_task: TaskID | None = None

    def started(self) -> None:
        self.progress = self.stack.enter_context(
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                MofNCompleteColumn(),
                TimeElapsedColumn(),
                TimeRemainingColumn(),
            ),
        )
        self.training_task = self.progress.add_task(
            "🔮 [cyan]Training speculator...[/cyan]",
            total=self.subsample_size,
        )

    def training_progress(self, trained_tokens: int) -> None:
        assert self.training_task is not None
        self.progress.update(self.training_task, completed=trained_tokens)

    def finished_training(self) -> None:
        assert self.training_task is not None
        self.progress.update(self.training_task, description="✅ Completed")
        self.progress.remove_task(self.training_task)
        self.stack.close()

    def saving_speculator(self) -> None:
        pass

    def finished_saving_speculator(self) -> None:
        console.print(f"💾 Speculator saved to [cyan]{self.output_path}[/cyan]")


@speculator_app.command(help="Train a speculator from inference traces")
def train(
    trace_path: Annotated[
        Path,
        Argument(
            help="File of llm inference traces to train the speculator on",
            metavar="TRACE_PATH",
        ),
    ],
    output_path: Annotated[
        Path,
        Option(
            help="File to save the output to",
            metavar="OUTPUT_PATH",
        ),
    ],
    hashtable_size: Annotated[
        int,
        Option(help="Size of ngram hashtable"),
    ] = 65536,
    num_logits_per_token: Annotated[
        int,
        Option(help="Top K tokens to keep in ngram hashtable"),
    ] = 8,
    ngram_size: Annotated[
        int,
        Option(help="Length of ngrams"),
    ] = 2,
    subsample_size: Annotated[
        int | None,
        Option(
            help="Exit early after training the model on this number of tokens",
            show_default="all",
        ),
    ] = None,
) -> None:
    _train(
        trace_path,
        output_path,
        hashtable_size,
        num_logits_per_token,
        ngram_size,
        subsample_size,
        CliTrainCallbacks,
    )


@speculator_app.command(help="Run speculator as an autoregressive llm")
def test(
    speculator_path: Annotated[
        Path,
        Argument(
            help="Path to the speculator file.",
            metavar="SPECULATOR_PATH",
        ),
    ],
    model_path: Annotated[
        Path,
        Argument(
            help="Path to the model directory for detokenization.",
            metavar="MODEL_PATH",
        ),
    ],
    seed: Annotated[
        int | None,
        Option(help="Set seed for deterministic sampling"),
    ] = None,
    num_sequences: Annotated[
        int,
        Option(help="Number of sequences to generate"),
    ] = 8,
) -> None:
    model = LanguageModelConfig.load_model(model_path)

    with open(speculator_path, "rb") as fd:
        speculator = NGramSpeculator.deserialize(fd.read())

    table = Table(
        show_header=False,
        show_lines=True,
        box=box.ROUNDED,
    )

    if seed is not None:
        random.seed(seed)

    for _ in range(num_sequences):
        sequence = test_speculator(speculator)
        detokenized = model.message_processor.detokenize(sequence)
        table.add_row(detokenized)

    console.print(table)


@app.callback()
def _profile_memory(
    ctx: Context,
    profile_memory: Annotated[
        Path | None,
        Option(
            help="Record and save the XLA memory profile to specified path",
            show_default="Don't save the XLA memory profile",
            envvar="LALAMO_PROFILE_MEMORY",
        ),
    ] = None,
) -> None:
    if profile_memory is None:
        return

    if profile_memory.is_dir():
        profile_memory /= "lalamo-memory.prof"

    def _save_memory_profile() -> None:
        console.print(f"Saving XLA memory profile to {profile_memory}")
        jax.profiler.save_device_memory_profile(profile_memory)

    ctx.call_on_close(_save_memory_profile)


if __name__ == "__main__":
    app()



from .common import State, StateLayerBase
from .kv_cache import DynamicKVCacheLayer, KVCacheLayer, StaticKVCacheLayer
from .mamba_state import Mamba2StateLayer
from .short_conv_state import ShortConvStateLayer

__all__ = [
    "DynamicKVCacheLayer",
    "KVCacheLayer",
    "Mamba2StateLayer",
    "ShortConvStateLayer",
    "State",
    "StateLayerBase",
    "StaticKVCacheLayer",
]



from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

import jax
from jax import Array
from jax import numpy as jnp
from jaxtyping import Float

from lalamo.message_processor import Message, MessageProcessor
from lalamo.modules import Classifier, ClassifierConfig, LalamoModule

from .common import TextModel, TextModelConfig

__all__ = [
    "ClassifierModel",
    "ClassifierModelConfig",
]


@dataclass(frozen=True)
class ClassifierModelConfig(TextModelConfig[ClassifierConfig]):
    def init(
        self,
        model: LalamoModule,
        message_processor: MessageProcessor,
    ) -> "ClassifierModel":
        assert isinstance(model, Classifier)
        return ClassifierModel(self, model, message_processor)

    @classmethod
    def load_model(cls, path: Path | str) -> "ClassifierModel":
        result = super().load_model(path)
        assert isinstance(result, ClassifierModel)
        return result


class ClassifierModel(TextModel[ClassifierModelConfig, Classifier]):
    def label_output_logits(self, logits: Float[Array, "batch logits"]) -> dict[str, Float[Array, " batch"]]:
        output_labels = self.model.config.output_labels
        probabilities = jax.nn.sigmoid(logits)

        if output_labels is None:
            output_labels = [f"class_{idx}" for idx in range(self.model.config.num_labels)]

        assert probabilities.ndim == 2, f"Expected 2D array, got array of shape {logits.shape}"

        return dict(zip(output_labels, jnp.unstack(probabilities, axis=1), strict=True))

    def classify_chat(
        self,
        messages: Iterable[Message],
    ) -> dict[str, float]:
        token_ids = jnp.array(self.message_processor.tokenize_request(messages), dtype=jnp.int32)[None, :]
        _, sequence_length = token_ids.shape
        token_positions = jnp.arange(sequence_length, dtype=jnp.int32)[None, :]
        classifier_output = self.model(token_ids=token_ids, token_positions=token_positions)

        return {k: float(v.item()) for k, v in self.label_output_logits(classifier_output.logits).items()}
